"""obs."""

import json
import logging
import os

import numpy as np

try:
    import requests
except ModuleNotFoundError:
    logging.debug("requests not loaded")
try:
    import xarray as xr
except ModuleNotFoundError:
    logging.debug("xarray not loaded")
try:
    import cfunits
except ModuleNotFoundError:
    cfunits = None
except AssertionError:
    cfunits = None
except:  # noqa
    cfunits = None


from .datetime_utils import as_datetime, as_datetime_args, as_timedelta, utcfromtimestamp
from .geo import LonLatVal
from .observation import Observation
from .titan import dataset_from_file


class ObservationSet(object):
    """Set of observations."""

    def __init__(self, observations, label="", sigmao=None):
        """Create an observation set.

        Args:
            observations (list): Observation objects.
            label (str, optional): Name of set. Defaults to "".
            sigmao (float, optional): Observation error relative to normal background
                                      error. Defaults to None.

        """
        self.size = len(observations)
        self.label = label
        self.index_pos = {}
        self.index_stid = {}

        if sigmao is not None:
            logging.info(
                "Setting sigmao=%s for all observations for label=%s", sigmao, label
            )
            for obs in observations:
                obs.sigmao = sigmao
        self.observations = observations

    def get_stid_index(self, stid):
        """Get station ID index.

        Args:
            stid (str): Station ID.

        Returns:
            int: Found index

        """
        stid = str(stid)
        if stid in self.index_stid:
            return self.index_stid[stid]
        return None

    def get_pos_index(self, lon, lat):
        """Get position index.

        Args:
            lon (float):Longitude
            lat (float): Latitude

        Returns:
            int: Found position index.
        """
        lon = Observation.format_lon(lon)
        lat = Observation.format_lat(lat)
        pos = lon + ":" + lat
        if pos in self.index_pos:
            return self.index_pos[pos]
        return None

    def get_obs(self):
        """Get observations.

        Returns:
            (list, list , list, list, list, list, list): times, lons, lats, stids, elevs,
                                                         values, varnames, sigmaos
        """
        obs2vectors = np.vectorize(Observation.obs2vectors)
        logging.debug("Obs dim %s", len(self.observations))
        if len(self.observations) > 0:
            times, lons, lats, stids, elevs, values, varnames, sigmaos = obs2vectors(
                self.observations
            )

            for point, lonval in enumerate(lons):
                lon = Observation.format_lon(lonval)
                lat = Observation.format_lat(lats[point])
                stid = str(stids[point])

                pos = lon + ":" + lat
                self.index_pos.update({pos: point})
                if stid != "NA":
                    self.index_stid.update({stid: point})

            times = times.tolist()
            lons = lons.tolist()
            lats = lats.tolist()
            stids = stids.tolist()
            elevs = elevs.tolist()
            values = values.tolist()
            varnames = varnames.tolist()
            sigmaos = sigmaos.tolist()

            return times, lons, lats, stids, elevs, values, varnames, sigmaos
        return [], [], [], [], [], [], [], []

    def matching_obs(self, my_obs):
        """Match the observations.

        Args:
            my_obs (Observation): Observation to match.

        Returns:
            bool: True if found

        """
        found = False
        for i in range(len(self.observations)):
            if my_obs.obstime == self.observations.obstimes[i]:
                lon = self.observations.obstimes[i]
                lat = self.observations.obstimes[i]
                pos = Observation.format_lon(lon) + ":" + Observation.format_lat(lat)

                if pos in self.index_pos:
                    found = True
        return found

    def points(self, geo, validtime=None):
        """Extract points from observations.

        Args:
            geo (surfex.Geo): Surfex geometry
            validtime (datetime.datetime, optional): Valid time. Defaults to None.

        Returns:
            np.ndarray: Values in points.
        """
        my_times = []
        my_values = []
        my_stids = []
        times, lons, lats, stids, __, values, __, __ = self.get_obs()

        for i in range(geo.nlons):
            lon = geo.lonlist[i]
            lat = geo.latlist[i]
            pos = Observation.format_lon(lon) + ":" + Observation.format_lat(lat)

            lons.append(lon)
            lats.append(lat)
            if pos in self.index_pos:
                ind = self.index_pos[pos]
                if validtime is not None:
                    logging.warning("No time check implemented yet")
                my_times.append(times[ind])
                my_stids.append(stids[ind])
                my_values.append(values[ind])
            else:
                my_times.append(None)
                my_stids.append("NA")
                my_values.append(np.nan)
                logging.info("Could not find position %s in this data source", pos)

        my_values = np.asanyarray(my_values)
        return my_times, my_values, my_stids

    def write_json_file(self, filename, indent=None):
        """Write a json file.

        Args:
            filename (str): Name of file
            indent (int, optional): Indentation in file. Defaults to None.
        """
        obs2vectors = np.vectorize(Observation.obs2vectors)
        data = {}
        if len(self.observations) > 0:
            obstimes, lons, lats, stids, elevs, values, varnames, sigmao = obs2vectors(
                self.observations
            )
            for obs, lon in enumerate(lons):
                data.update(
                    {
                        obs: {
                            "obstime": obstimes[obs].strftime("%Y%m%d%H%M%S"),
                            "varname": varnames[obs],
                            "lon": lon,
                            "lat": lats[obs],
                            "stid": stids[obs],
                            "elev": elevs[obs],
                            "value": values[obs],
                            "sigmao": sigmao[obs],
                        }
                    }
                )
        with open(filename, mode="w", encoding="utf-8") as file_handler:
            json.dump(data, file_handler, indent=indent)

    def get_data_set(self, varname):
        """Get data set.

        Args:
            varname (_type_): _description_

        Returns:
            _type_: _description_
        """
        data = {}
        posids = []
        locations = {}
        times = []
        ctimes = []
        lons = {}
        lats = {}
        elevs = {}
        for obs in self.observations:
            if obs.varname == varname:
                time = obs.obstime
                logging.info("time: %s", time)
                ctime = f"{time.strftime('%Y%m%d%H%M')}"
                posid = obs.posid
                stid = obs.stid
                if stid != "NA":
                    posid = stid

                posid = posid.replace("SN", "")
                posid = int(posid)

                ident = f"{time.strftime('%Y%m%d%H%M')}#{posid}"
                if ctime not in ctimes:
                    ctimes.append(ctime)
                    times.append(time)

                if posid not in posids:
                    posids.append(posid)

                stid = obs.stid
                if stid == "NA":
                    stid = posid
                locations.update({posid: stid})
                data.update({ident: obs.value})
                lons.update({posid: obs.lon})
                lats.update({posid: obs.lat})
                elevs.update({posid: obs.elev})

        ntimes = len(times)
        nlocations = len(posids)
        values = np.nan * np.zeros(
            [
                ntimes,
                nlocations,
            ],
            np.float32,
        )

        dtimes = np.nan * np.zeros([ntimes], np.double)
        dlons = np.nan * np.zeros([nlocations], np.float32)
        dlats = np.nan * np.zeros([nlocations], np.float32)
        delevs = np.nan * np.zeros([nlocations], np.float32)
        dtimes = np.array(times).astype("datetime64[s]").astype("int")
        for t, time in enumerate(times):
            ctime = time.strftime("%Y%m%d%H%M")
            for pos, posid in enumerate(posids):
                ident = f"{ctime}#{posid}"
                values[t][pos] = data[ident]
                dlons[pos] = lons[posid]
                dlats[pos] = lats[posid]
                delevs[pos] = elevs[posid]

        logging.info("posids %s", posids)
        coords = {}

        dtimes = np.array(times)
        coords["obstime"] = (["obstime"], [], {}, {"dtype": np.datetime64})
        coords["location"] = (["location"], posids)
        coords["lat"] = (
            ["location"],
            dlats,
            {"units": "degrees_east"},
        )
        coords["lon"] = (
            ["location"],
            dlons,
            {"units": "degrees_north"},
        )
        coords["altitude"] = (
            ["location"],
            delevs,
            {"units": "m"},
        )

        ds = xr.Dataset(coords=coords)
        ds["obstime"] = dtimes
        ds["obs"] = (["obstime", "location"], values)
        return ds


class NetatmoObservationSet(ObservationSet):
    """Observation set from netatmo."""

    def __init__(
        self,
        filenames,
        variable,
        target_time,
        dt=3600,
        re=True,
        lonrange=None,
        latrange=None,
        label="netatmo",
        sigmao=None,
    ):
        """Construct netatmo obs.

        Args:
            filenames (list): Filenames
            variable (str): Variable
            target_time (as_datetime): _description_
            dt (int, optional): _description_. Defaults to 3600.
            re (bool, optional): _description_. Defaults to True.
            lonrange (_type_, optional): _description_. Defaults to None.
            latrange (_type_, optional): _description_. Defaults to None.
            label (str, optional): _description_. Defaults to "netatmo".
            sigmao (float, optional): Observation error relative to normal background
                                      error. Defaults to None.

        Raises:
            RuntimeError: Lonrange must be a list with length 2
            RuntimeError: Latrange must be a list with length 2

        """
        if lonrange is None:
            lonrange = [-180, 180]

        if not isinstance(lonrange, list) or len(lonrange) != 2:
            raise RuntimeError(f"Lonrange must be a list with length 2 {lonrange}")

        if latrange is None:
            latrange = [-90, 90]

        if not isinstance(latrange, list) or len(latrange) != 2:
            raise RuntimeError(f"Latrange must be a list with length 2 {latrange}")

        data = {}  # key: id, value: list of values
        times = {}  # key: id, value: list of times
        metadata = {}

        # Parse json data
        #
        # The raw data is not valid JSON, since it is missing commas between lists
        # e.g. [...][...][...]. Instead format it like this: [..., ..., ...]

        observations = []
        num_missing_metadata = 0
        num_missing_obs = 0
        num_missing_time = 0
        num_missing_elev = 0
        num_wrong_time = 0
        for ifilename in filenames:
            with open(ifilename, mode="r", encoding="utf-8") as ifile:
                text = ifile.read()

            try:
                # Older files had an additional bug. This code will work for both cases
                if len(text) == 0:
                    # Sometimes netatmo files are empty
                    logging.info("Empty file: %s", ifilename)
                    continue

                if text[0] == "{":
                    text = f"[{text}"

                text = text.replace("}]{", "}{")
                text = text.replace("}][{", "},{")
                text = text.replace("}{", "},{")
                text = '{"data": %s}' % text
                raw = json.loads(text)
                raw = raw["data"]
                logging.debug("Parsing %d stations in %s", len(raw), ifilename)
            except RuntimeError:
                logging.error("Could not parse %s.", ifilename)
                continue

            for line in raw:
                if "data" in line and "_id" in line and "location" in line:
                    my_id = line["_id"]
                    location = line["location"]
                    curr_data = line["data"]
                    if variable in curr_data:
                        if "time_utc" in curr_data:
                            time_utc = curr_data["time_utc"]
                            # Record this observation if it is closer to the target
                            # time than any other previously parsed observation for
                            # this location. Also, only record if the time difference
                            # is within acceptable limits.
                            if "altitude" not in line:
                                num_missing_elev += 1

                            if not re or "altitude" in line:
                                lon = location[0]
                                lat = location[1]
                                if (
                                    lonrange[0] <= lon <= lonrange[1]
                                    and latrange[0] <= lat <= latrange[1]
                                ):
                                    if my_id not in data:
                                        data[my_id] = []
                                        times[my_id] = []

                                        elev = np.nan
                                        metadata[my_id] = {
                                            "lon": lon,
                                            "lat": lat,
                                            "elev": elev,
                                        }
                                    if (
                                        np.isnan(metadata[my_id]["elev"])
                                        and "altitude" in line
                                    ):
                                        metadata[my_id]["elev"] = line["altitude"]

                                    value = curr_data[variable]
                                    if variable == "Temperature":
                                        value = value + 273.15
                                    if variable == "Humidity":
                                        value = value * 0.01
                                    data[my_id] += [value]
                                    times[my_id] += [time_utc]
                        else:
                            num_missing_time += 1
                    else:
                        num_missing_obs += 1
                else:
                    num_missing_metadata += 1

        if target_time is not None:
            num_valid_stations = 0
            for my_id, time in times.items():
                this_diff_times = [
                    (utcfromtimestamp(t) - target_time).total_seconds() for t in time
                ]
                curr_times = [utcfromtimestamp(t) for t in time]
                if np.min(np.abs(np.array(this_diff_times))) < dt:
                    ibest = int(np.argmin(np.abs(np.array(this_diff_times))))
                    curr_time = curr_times[ibest]
                    elev = metadata[my_id]["elev"]
                    observations.append(
                        Observation(
                            curr_time,
                            metadata[my_id]["lon"],
                            metadata[my_id]["lat"],
                            data[my_id][ibest],
                            elev=elev,
                            varname=variable,
                        )
                    )
                    num_valid_stations += 1
        else:
            num_valid_stations = len(data)

        logging.info("Found %d valid observations:", num_valid_stations)
        logging.info("   %d missing obs", num_missing_obs)
        logging.info("   %d missing metadata", num_missing_metadata)
        logging.info("   %d missing timestamp", num_missing_time)
        logging.info("   %d wrong timestamp", num_wrong_time)
        extra = " (not removed)" if not re else ""
        logging.debug("   %d missing elev%s", num_missing_elev, extra)

        ObservationSet.__init__(self, observations, label=label, sigmao=sigmao)


class MetFrostObservations(ObservationSet):
    """Observations from MET-Norway obs API (frost)."""

    def __init__(
        self,
        varname,
        stations=None,
        level=None,
        num_tries=3,
        wmo=None,
        providers=None,
        xproviders=None,
        blacklist=None,
        validtime=None,
        dt=3600,
        lonrange=None,
        latrange=None,
        unit=None,
        label="frost",
        sigmao=None,  # noqa ARG002
    ):
        """Construct obs set from Frost.

        Args:
            varname (_type_): _description_
            stations (_type_, optional): _description_. Defaults to None.
            level (_type_, optional): _description_. Defaults to None.
            num_tries (int, optional): _description_. Defaults to 3.
            wmo (_type_, optional): _description_. Defaults to None.
            providers (_type_, optional): _description_. Defaults to None.
            xproviders (_type_, optional): _description_. Defaults to None.
            blacklist (_type_, optional): _description_. Defaults to None.
            validtime (_type_, optional): _description_. Defaults to None.
            dt (int, optional): _description_. Defaults to 3600.
            lonrange (_type_, optional): _description_. Defaults to None.
            latrange (_type_, optional): _description_. Defaults to None.
            unit (_type_, optional): _description_. Defaults to None.
            label (str, optional): _description_. Defaults to "frost".
            sigmao (float, optional): Observation error relative to normal
                                      background error. Defaults to None.

        Raises:
            KeyError: _description_
            Exception: _description_
            Exception: _description_
            Exception: _description_
            Exception: _description_

        """
        if blacklist is None:
            blacklist = []

        # extract client ID from environment variable
        if "CLIENTID" not in os.environ:
            raise KeyError("error: CLIENTID not found in environment\n")

        client_id = os.environ["CLIENTID"]

        # Get all the stations (in an area or country)
        # Make list of station IDs and dictionary of their lat,long,elev
        #
        ids = []
        tries = 1
        station_dict = {}
        while tries <= num_tries:
            tries += 1
            # first get all the stations within a polygon or in a country?
            # 'geometry': 'POLYGON((10 60, 10 59, 11 60, 11 59))' # area around Oslo
            # 'country': 'NO' # everything in Norway

            parameters = {
                "types": "SensorSystem",
                "fields": "id,geometry,masl,wmoid,stationholders",
            }
            if lonrange is not None and latrange is not None:
                parameters.update(
                    {
                        "geometry": "POLYGON(("
                        + str(lonrange[0])
                        + " "
                        + str(latrange[0])
                        + ", "
                        + str(lonrange[0])
                        + " "
                        + str(latrange[1])
                        + ", "
                        + str(lonrange[1])
                        + " "
                        + str(latrange[1])
                        + ", "
                        + str(lonrange[1])
                        + " "
                        + str(latrange[0])
                        + " ))"
                    }
                )
            logging.debug("Request parameters: %s", str(parameters))
            req = requests.get(
                "https://frost.met.no/sources/v0.jsonld",
                parameters,
                auth=(client_id, ""),
                timeout=30,
            )

            logging.debug(
                "Request https://frost.met.no/sources/v0.jsonld returned %s",
                req.status_code,
            )
            # extract list of stations (if response was valid)
            if req.status_code == 200:
                data = req.json()["data"]
                ids = []
                count_discard = 0
                for data_block in data:
                    my_id = data_block["id"]
                    elev = data_block.get("masl", -999)

                    # filter data for WMO and non WMO
                    keep_this_id = True
                    if "wmoId" in data_block and wmo is not None and wmo == 0:
                        # station is WMO skip
                        keep_this_id = False
                        logging.debug("throwing out this id (is WMO): %s", my_id)
                    elif "wmoId" not in data_block and wmo is not None and wmo == 1:
                        # station is not WMO skip
                        keep_this_id = False
                        logging.debug("throwing out this id (not WMO): %s", my_id)

                    # filter out stations with incomplete data
                    if keep_this_id and "geometry" not in data_block:
                        keep_this_id = False
                        logging.debug("throwing out this id (no geometry): %s", my_id)

                    # filters for station holders
                    if "stationHolders" not in data_block:
                        keep_this_id = False
                        logging.debug(
                            "throwing out this id (no stationHolders): %s", my_id
                        )
                    # select station providers
                    elif providers is not None:
                        providers = providers.split(",")
                        station_holders = data_block["stationHolders"]
                        if not any(x in station_holders for x in providers):
                            keep_this_id = False
                            logging.debug(
                                "throwing out this id (station holder): %s",
                                str(station_holders),
                            )
                    # or exclude certain station providers
                    elif xproviders is not None:
                        xproviders = xproviders.split(",")
                        station_holders = data_block["stationHolders"]
                        if any(x in station_holders for x in xproviders):
                            keep_this_id = False
                            logging.debug(
                                "throwing out this id (exclude station holder): %s",
                                str(station_holders),
                            )

                    # filter out blacklisted stations
                    if my_id in blacklist:
                        keep_this_id = False
                        logging.debug("throwing out blacklisted id: %s", my_id)

                    if stations is not None and my_id not in stations:
                        keep_this_id = False
                        logging.debug(
                            "Throwing out station because not in station list %s",
                            my_id,
                        )

                    logging.debug(
                        "Keep this ID: %s bool: %s", str(my_id), str(keep_this_id)
                    )
                    if keep_this_id:  # write into dict
                        ids.append(my_id)
                        # create a dictionary for these stations to store
                        # lat,long,elev for each
                        station_dict[my_id] = [
                            data_block["geometry"]["coordinates"][1],
                            data_block["geometry"]["coordinates"][0],
                            elev,
                        ]
                    else:
                        count_discard = count_discard + 1
                logging.debug("Number of stations: %s", str(len(ids)))
                logging.debug(
                    "Number of stations , debug=debugdiscarded: %s", str(count_discard)
                )
                break

            if req.status_code == 404:
                logging.warning("STATUS: No data was found for the list of query Ids.")
                break
            if tries > num_tries:
                raise Exception("ERROR: could not retrieve observations.")

        #
        # Use the station ID list to get the observation for each station
        #
        ids_obs_dict = {}  # declare outside loop, since may be more than one request
        # check how long the list of stations is and potentially break it up to shorten
        observations = []
        it_ids = len(ids)
        dt = as_timedelta(seconds=dt)
        while it_ids > 0:
            if it_ids > 50:
                # get last 50
                sub_id_list = ids[it_ids - 50 : it_ids]
                it_ids = it_ids - 50
            else:
                # get the rest if <50
                sub_id_list = ids[:it_ids]
                it_ids = 0

            tries = 1
            while tries <= num_tries:
                tries += 1
                # use the list of stations and get the observations for those
                parameters2 = {"sources": ",".join(sub_id_list), "elements": varname}
                date = validtime.strftime("%Y%m%d")
                hour = validtime.strftime("%H")

                # if have specified a date and time
                if date is not None and hour is not None:
                    # make these into a format that works for FROST
                    date_string = date
                    hour_string = hour
                    date_string_frost = (
                        date_string[0:4]
                        + "-"
                        + date_string[4:6]
                        + "-"
                        + date_string[6:8]
                        + "T"
                        + hour_string
                    )
                    parameters2["referencetime"] = date_string_frost
                # do not have date and time, so use latest
                else:
                    parameters2["referencetime"] = "latest"
                    parameters2["maxage"] = "PT30M"
                    parameters2["limit"] = 1

                logging.debug("Request parameters2: %s", str(parameters2))
                req = requests.get(
                    "https://frost.met.no/observations/v0.jsonld",
                    parameters2,
                    auth=(client_id, ""),
                    timeout=30,
                )

                logging.debug(
                    "Request https://frost.met.no/observations/v0.jsonld returned %s",
                    req.status_code,
                )
                if req.status_code == 200:
                    data = req.json()["data"]
                    for data_block in data:
                        # Check that reference time is ok, since sometimes
                        # future observations can be present when 'latest' is
                        # chosen for reference time
                        ref_str = data_block["referenceTime"]
                        ref_year = int(ref_str[0:4])
                        ref_month = int(ref_str[5:7])
                        ref_day = int(ref_str[8:10])
                        ref_hour = int(ref_str[11:13])
                        ref_min = int(ref_str[14:16])
                        ref_sec = int(ref_str[17:19])
                        ref_time = as_datetime_args(
                            year=ref_year,
                            month=ref_month,
                            day=ref_day,
                            hour=ref_hour,
                            minute=ref_min,
                            second=ref_sec,
                        )
                        logging.debug("ref_time %s validtime %s", ref_time, validtime)

                        read_unit = None
                        levels_ok = True
                        if "observations" in data_block:
                            for obs in data_block["observations"]:
                                logging.debug("%s", obs)
                                if "unit" in obs:
                                    read_unit = obs["unit"]
                                if "level" in obs and level is not None:
                                    logging.debug("level %s", obs["level"])
                                    all_found = True
                                    for key in level:
                                        if key in obs["level"]:
                                            if str(level[key]) != str(obs["level"][key]):
                                                logging.debug(
                                                    "%s != %s",
                                                    level[key],
                                                    obs["level"][key],
                                                )
                                                all_found = False
                                            else:
                                                logging.debug(
                                                    "%s == %s",
                                                    level[key],
                                                    obs["level"][key],
                                                )
                                    if not all_found:
                                        levels_ok = False

                        keep_this_obs = False
                        if levels_ok and abs(ref_time - validtime) < dt:
                            keep_this_obs = True

                        if keep_this_obs:
                            logging.debug("Keep this obs")
                            value = data_block["observations"][0]["value"]
                            if len(str(value)) > 0:  # not all stations have observations
                                source_id = str(data_block["sourceId"])
                                my_id = source_id.split(":")
                                if unit is not None:
                                    if read_unit is not None:
                                        if cfunits is None:
                                            raise Exception("cfunits not loaded!")
                                        read_unit = cfunits.Units(read_unit)
                                        unit = cfunits.Units(unit)
                                        value = cfunits.Units.conform(
                                            value, read_unit, unit
                                        )
                                    else:
                                        raise Exception("Did not read a unit to convert!")
                                ids_obs_dict[my_id[0]] = value

                    logging.debug(
                        "Station list length: %s, total number of observations "
                        "retrieved: %s",
                        str(len(sub_id_list)),
                        str(len(ids_obs_dict)),
                    )
                    break
                if req.status_code == 404:
                    logging.warning(
                        "STATUS: No data was found for the list of query Ids."
                    )
                    break
                if tries > num_tries:
                    raise RuntimeError("ERROR: could not retrieve observations.")

            for station, station_id in ids_obs_dict.items():
                value = float(station_id)
                id_info = station_dict[station]
                stid = str(station)[2:]
                lat = id_info[0]
                lon = id_info[1]
                elev = id_info[2]
                observations.append(
                    Observation(
                        validtime,
                        lon,
                        lat,
                        value,
                        stid=str(stid),
                        elev=elev,
                        varname=varname,
                    )
                )

        ObservationSet.__init__(self, observations, label=label)


class JsonObservationSet(ObservationSet):
    """JSON observation set."""

    def __init__(self, filename, label="json", var=None, sigmao=None):
        """Construct an observation data set from a json file.

        Args:
            filename (str): Filename
            label (str, optional): Label of set. Defaults to "json".
            var (str, optional): Variable name. Defaults to None.
            sigmao (float, optional): Observation error relative to normal background
                                      error. Defaults to None.

        Raises:
            RuntimeError: Varname is not found

        """
        with open(filename, mode="r", encoding="utf-8") as file_handler:
            obs = json.load(file_handler)

        observations = []
        for i in range(len(obs)):
            ind = str(i)
            obstime = as_datetime(obs[ind]["obstime"])
            lon = obs[ind]["lon"]
            lat = obs[ind]["lat"]
            elev = obs[ind]["elev"]
            value = obs[ind]["value"]
            stid = obs[ind]["stid"]
            if sigmao is not None:
                lsigmao = sigmao
            else:
                try:
                    lsigmao = obs[ind]["sigmao"]
                except KeyError:
                    lsigmao = 1.0
            varname = ""
            if "varname" in obs[ind]:
                varname = obs[ind]["varname"]

            if varname == "" and var is not None:
                raise RuntimeError("Varname is not found " + varname)

            if var is None or var == varname:
                observations.append(
                    Observation(
                        obstime,
                        lon,
                        lat,
                        value,
                        stid=stid,
                        elev=elev,
                        varname=varname,
                        sigmao=lsigmao,
                    )
                )

        ObservationSet.__init__(self, observations, label=label, sigmao=sigmao)


class ObservationFromTitanJsonFile(ObservationSet):
    """Observation set from titan json file."""

    def __init__(self, an_time, filename, label="", sigmao=None):
        """Constuct obs set from a titan json file.

        Args:
            an_time (_type_): _description_
            filename (_type_): _description_
            label (str, optional): _description_. Defaults to "".
            sigmao (float, optional): Observation error relative to normal background
                                      error. Defaults to None.

        """
        qc_obs = dataset_from_file(an_time, filename)
        observations = []
        for i in range(len(qc_obs)):
            observations.append(
                Observation(
                    qc_obs.obstimes[i],
                    qc_obs.lons[i],
                    qc_obs.lats[i],
                    qc_obs.elevs[i],
                    qc_obs.values[i],
                    sigmao=qc_obs.epsilons[i],
                )
            )

        ObservationSet.__init__(self, observations, label=label, sigmao=sigmao)


class ObsSetFromVobs(ObservationSet):
    """Create observation set from vobs files."""

    def __init__(self, fname, validtime, varname=None, label="vobs", sigmao=None):
        """Constuct obs set from vobs file.

        Args:
            fname (_type_): File name
            validtime (_type_): Valid time
            varname (str, optional): Variable name. Defaults to "".
            label (str, optional): _description_. Defaults to "vobs".
            sigmao (float, optional): Observation error relative to normal background
                                      error. Defaults to None.

        """
        logging.info("Processing %s", fname)
        observations = []
        obpars, obsx = self.read_vfld(fname)

        for obname in obpars:
            for stid, odata in obsx.items():
                lon = odata["lon"]
                lat = odata["lat"]
                elev = odata["hgt"]
                obs = np.nan
                if obname in obsx[stid]:
                    obs = obsx[stid][obname]
                if obs == -99:
                    obs = np.nan
                value = obsx[stid][obname]
                if varname is None or varname == obname:
                    observations.append(
                        Observation(
                            validtime,
                            lon,
                            lat,
                            value,
                            elev=elev,
                            stid=stid,
                            varname=obname,
                        )
                    )

        ObservationSet.__init__(self, observations, label=label, sigmao=sigmao)

    @staticmethod
    def read_vfld(fnam):
        """Read vfld file."""
        with open(fnam, mode="r", encoding="utf-8") as f:
            h1 = f.readline().split()
            h2 = f.readline()
            nst, __, __ = (int(i) for i in h1)
            npar = int(h2)
            pars = {}
            for i in range(npar):
                l = f.readline().split()  # noqa
                pars[l[0]] = {"index": i, "val": int(l[1])}
            extra = 3 if "FI" in pars else 4
            x = {}
            for i in range(nst):  # noqa
                l = f.readline().split()  # noqa
                lat, lon, hgt = l[1:4]
                stnr = l[0][-5:]
                x[stnr] = {"lat": float(lat), "lon": float(lon), "hgt": float(hgt)}
                for par, val in pars.items():
                    if par != "FI":
                        x[stnr][par] = float(l[val["index"] + extra])

        return pars, x


class StationList:
    """Station list."""

    def __init__(self, fname):
        """Construct."""
        self.fname = fname
        index_pos = {}
        aliases = {}
        try:
            with open(fname, mode="r", encoding="utf-8") as file_handler:
                ids_from_file = json.load(file_handler)
        except FileNotFoundError:
            raise FileNotFoundError("Station list does not exist!") from FileNotFoundError
        self.stids = ids_from_file
        ids = []
        for stid in ids_from_file:
            ids.append(stid)
            lon = ids_from_file[stid]["lon"]
            lat = ids_from_file[stid]["lat"]
            pos = self.posid(lon, lat)
            index_pos.update({pos: stid})
            if "aliases" in ids_from_file[stid]:
                aliases.update({pos: ids_from_file[stid]["aliases"]})
        self.ids = ids
        self.index_pos = index_pos
        self.geo = self.get_geo()

    def get_geo(self, lonrange=None, latrange=None, xdx="0.3", xdy="0.3"):
        """Set geometry from station list."""
        if lonrange is None:
            lonrange = [-180, 180]
        if latrange is None:
            latrange = [-90, 90]

        lons = []
        lats = []
        for stid in self.stids:
            lon, lat, __ = self.get_pos_from_stid([stid])
            lon = lon[0]
            lat = lat[0]
            if lonrange[0] <= lon <= lonrange[1] and latrange[0] <= lat <= latrange[1]:
                lon = round(lon, 5)
                lat = round(lat, 5)
                lons.append(lon)
                lats.append(lat)

        d_x = [xdx] * len(lons)
        d_y = [xdy] * len(lats)
        geo_json = {
            "nam_pgd_grid": {"cgrid": "LONLATVAL"},
            "nam_lonlatval": {"xx": lons, "xy": lats, "xdx": d_x, "xdy": d_y},
        }
        return LonLatVal(geo_json)

    def get_pos_from_stid(self, stids):
        """Get pos from station ID.

        Args:
            stids (list): _description_

        Raises:
            RuntimeError: _description_

        Returns:
            tuple: longitudes, latitudes

        """
        lons = []
        lats = []
        elevs = []
        if isinstance(stids, str):
            stids = [stids]

        found = False
        for stid in self.stids:
            for stid1 in stids:
                if stid == stid1:
                    found = True
                    lon = float(self.stids[stid1]["lon"])
                    lat = float(self.stids[stid1]["lat"])
                    elev = "NA"
                    if "elev" in self.stids[stid1]:
                        elev = self.stids[stid1]["elev"]
                    lons.append(lon)
                    lats.append(lat)
                    if elev == "NA":
                        elev = np.nan
                    elevs.append(elev)
        if not found:
            raise RuntimeError(
                "Could not find station id " + stid + " in file " + self.fname
            )
        return lons, lats, elevs

    def get_stid_from_stationlist(self, lons, lats):
        """Get station ID from station list.

        Args:
            lons (list): Longitudes
            lats (list): Latitudes

        Returns:
            list: Station IDs

        """
        stids = []
        for i, lon in enumerate(lons):
            lat = lats[i]
            pos = self.posid(lon, lat)
            if pos in self.index_pos:
                stids.append(self.index_pos[pos])
            else:
                stids.append("NA")
        return stids

    @staticmethod
    def posid(lon, lat, pos_decimals=5):
        """Return pos id.

        Args:
            lon (_type_): _description_
            lat (_type_): _description_
            pos_decimals (int, optional): _description_. Defaults to 5.

        Returns:
            _type_: _description_
        """
        return f"{lon:.{pos_decimals}f}#{lat:.{pos_decimals}f}"

    def posids(self, lons, lats, pos_decimals=5):
        """Get posids.

        Args:
            lons (_type_): _description_
            lats (_type_): _description_
            pos_decimals (int, optional): _description_. Defaults to 5.

        Returns:
            _type_: _description_
        """
        lposids = []
        for pind, lon in enumerate(lons):
            lat = lats[pind]
            pos = self.posid(lon, lat, pos_decimals=pos_decimals)
            if pos in self.index_pos:
                pos = self.index_pos[pos]
            lposids.append(pos)
        return lposids

    def all_posids(self, pos_decimals=5):
        """Get all posids in list.

        Args:
            pos_decimals (int, optional): _description_. Defaults to 5.

        Returns:
            _type_: _description_
        """
        lons, lats, __ = self.get_pos_from_stid(self.stids)
        return self.posids(lons, lats, pos_decimals=pos_decimals)



#!/usr/bin/env python3
"""pysurfex module."""
from importlib.metadata import version
from pathlib import Path

PACKAGE_NAME = __name__
__version__ = version(__name__)

PACKAGE_DIRECTORY = Path(__file__).parent



"""Verification."""

import json
import logging
import os
import sys
from argparse import ArgumentParser
from datetime import datetime

try:
    import pandas
except ModuleNotFoundError:
    logging.debug("pandas not loaded")

try:
    from grib2sqlite import create_table, sqlite_name, write_to_sqlite
except ModuleNotFoundError:
    sqlite_name = None
    create_table = None
    write_to_sqlite = None

import numpy as np

try:
    import xarray as xr
except ModuleNotFoundError:
    logging.debug("xarray not loaded")

from pysurfex.cmd_parsing import parse_args_variable, variable_parse_options
from pysurfex.datetime_utils import as_datetime, as_timedelta
from pysurfex.geo import ConfProj, get_geo_object
from pysurfex.obs import ObservationSet, ObsSetFromVobs, StationList
from pysurfex.observation import Observation
from pysurfex.read import ConvertedInput, get_multi_converters


class ObsDataFromSurfexConverter(ObservationSet):
    """Create verif data from pysurfex converter."""

    def __init__(
        self, converter, var, stationlist, validtime, cache=None, label="converter"
    ):
        """Construct obs dataset from a converter.

        Args:
            converter (_type_): _description_
            var (_type_): _description_
            stationlist (_type_): _description_
            validtime (_type_): _description_
            cache (_type_, optional): _description_. Defaults to None.
            label (str, optional): _description_. Defaults to "converter".

        """
        pvalues = ConvertedInput(stationlist.geo, var.name, converter).read_time_step(
            validtime, cache
        )

        validtime = datetime.strptime(validtime.strftime("%Y%m%d%H"), "%Y%m%d%H")
        observations = []
        lons, lats, elevs = stationlist.get_pos_from_stid(stationlist.stids)
        for ind, stid in enumerate(stationlist.stids):
            observations.append(
                Observation(
                    validtime,
                    lons[ind],
                    lats[ind],
                    pvalues[ind],
                    stid=stid,
                    elev=elevs[ind],
                    varname=var.name,
                )
            )

        ObservationSet.__init__(self, observations, label=label)
        self.ds = self.get_data_set(var.name)


class DatasetFromVfld:
    """Create observation set from vfld files."""

    def __init__(self, fname, basetime, validtime, variable, stationlist=None):
        """Construct a dataset from a vfld file.

        Args:
            fname (_type_): _description_
            basetime (_type_): _description_
            validtime (_type_): _description_
            variable (_type_): _description_
            stationlist (_type_, optional): _description_. Defaults to None.

        """
        pars, x = ObsSetFromVobs.read_vfld(fname)
        fcst = []
        basetime = datetime.strptime(basetime.strftime("%Y%m%d%H"), "%Y%m%d%H")
        validtime = datetime.strptime(validtime.strftime("%Y%m%d%H"), "%Y%m%d%H")
        time = [basetime]
        leadtime = [validtime - basetime]
        location = []
        # TODO
        del pars["FI"]
        for par in pars:
            if variable is None or par == variable:
                for stid, xdata in x.items():
                    # TODO check on stationlist
                    if stationlist is None:
                        location.append(int(stid))
                        # lon = xdata["lon"] # noqa
                        # lat = xdata["lat"] # noqa
                        # elev = xdata["hgt"] # noqa
                        value = xdata[par]
                        fcst.append(value)

        if len(location) > 0:
            fcst = np.array(fcst)
            fcst = fcst.reshape((1, 1, len(location)))
            ds = xr.Dataset(
                data_vars={"fcst": (["time", "leadtime", "location"], fcst)},
                coords={"time": time, "leadtime": leadtime, "location": location},
            )
        else:
            ds = xr.Dataset()
        self.ds = ds


class VerificationVariable:
    """Parent verification variable."""

    def __init__(self, name):
        """Construct a verification variable.

        Args:
            name (_type_): _description_
        """
        self.name = name


class VerifVariable(VerificationVariable):
    """Verif variable."""

    def __init__(self, name, unit=None):
        """Construct a verif variable.

        Args:
            name (_type_): _description_
            unit (_type_, optional): _description_. Defaults to None.
        """
        if unit is None:
            unit = ""
        self.unit = unit
        VerificationVariable.__init__(self, name)


class VerificationDataFromSurfexConverter:
    """Create verif data from pysurfex converter."""

    def __init__(self, converter, var, stationlist, validtime, basetime, cache=None):
        """Construct verification data from a converter.

        Args:
            converter (_type_): _description_
            var (_type_): _description_
            stationlist (_type_): _description_
            validtime (_type_): _description_
            basetime (_type_): _description_
            cache (_type_, optional): _description_. Defaults to None.

        """
        pvalues = ConvertedInput(stationlist.geo, var.name, converter).read_time_step(
            validtime, cache
        )

        basetime = converter.initial_time
        lons = stationlist.geo.lons
        nlons = len(lons)
        ileadtime2 = np.array([validtime - basetime]).astype(np.timedelta64)
        ileadtime = [np.timedelta64(elem, "h").astype("int") for elem in ileadtime2]
        cbasetime = np.array([basetime]).astype(np.datetime64)
        basetime = np.array(basetime).astype(np.datetime64)

        pvalues = pvalues.reshape(1, 1, nlons)

        posids = stationlist.all_posids()
        posids2 = []
        for pid in posids:
            pid2 = pid.replace("SN", "")
            posids2.append(int(pid2))
        cbasetime = np.array([basetime]).astype(np.datetime64)

        ds = xr.Dataset(
            data_vars={"fcst": (["time", "leadtime", "location"], pvalues)},
            coords={
                "time": (["time"], cbasetime),
                "leadtime": (["leadtime"], ileadtime),
                "location": (["location"], posids2),
            },
            attrs={"long_name": var.name, "units": var.unit},
        )
        self.ds = ds


class DataFromSurfexConverter:
    """Create verif data from pysurfex converter."""

    def __init__(self, converter, var, geo, validtime, cache=None):
        """Construct verification data from a converter.

        Args:
            converter (_type_): _description_
            var (_type_): _description_
            geo (_type_): _description_
            validtime (_type_): _description_
            basetime (_type_): _description_
            cache (_type_, optional): _description_. Defaults to None.

        """
        pvalues = ConvertedInput(geo, var.name, converter).read_time_step(
            validtime, cache
        )

        pvalues = pvalues.reshape([1, geo.nlats, geo.nlons], order="F")

        validtime = datetime.strptime(validtime.strftime("%Y%m%d%H"), "%Y%m%d%H")
        cvalidtime = np.array([validtime]).astype(np.datetime64)

        attrs = {"long_name": var.name, "units": var.unit}
        if isinstance(geo, ConfProj):
            attrs.update(
                {
                    "gridtype": "lambert",
                    "dlon": float(geo.xdx),
                    "dlat": float(geo.xdy),
                    "projlat": float(geo.xlat0),
                    "projlat2": float(geo.xlat0),
                    "projlon": float(geo.xlon0),
                    "lonc": float(geo.xloncen),
                    "latc": float(geo.xlatcen),
                }
            )
        data_vars = {
            f"{var.name}": (["time", "y", "x"], pvalues),
            "longitude": (["y", "x"], np.transpose(geo.lons)),
            "latitude": (["y", "x"], np.transpose(geo.lats)),
        }
        ds = xr.Dataset(
            data_vars=data_vars,
            coords={
                "time": (["time"], cvalidtime),
                "y": (["y"], geo.yyy),
                "x": (["x"], geo.xxx),
            },
            attrs=attrs,
        )
        self.ds = ds


class VerifData:
    """Verif data."""

    def __init__(self, ds, variable):
        """Construct VerifData.

        Args:
            ds (xr.Dataset): Dataset
            variable (str): Variable

        """
        self.var = variable
        self.ds = ds

    def merge(self, ds_model, ds_obs):
        """Merge model and obs to verif data.

        Args:
            ds_model (xr.Dataset): Model dataset
            ds_obs (xr.Dataset): Observation data set

        """
        logging.info("model: %s", ds_model)
        logging.info("obs: %s", ds_obs)

        leadtime = ds_model["leadtime"].data
        model_reference_time = ds_model["time"].data
        location = ds_model["location"].data
        model_valid_time = []
        for tim in model_reference_time:
            for lt in leadtime:
                model_valid_time.append(tim + lt)
        unique_model_valid_time = np.sort(np.unique(model_valid_time))
        ntimes = unique_model_valid_time.shape[0]

        # dummy observations
        nlocations = location.shape[0]
        dummy_obs = np.nan * np.random.randn(ntimes, nlocations)
        ds3 = xr.Dataset(
            data_vars={"obs": (["obstime", "location"], dummy_obs)},
            coords={
                "obstime": unique_model_valid_time,
                "location": location,
            },
        )
        ds_obs = ds3.update(ds_obs)

        fcst = ds_model.fcst.data
        obs = ds_obs.obs.sel(obstime=ds_model.time + ds_model.leadtime).data
        leadtime = ds_model.leadtime.data.astype("timedelta64[h]").astype("int")
        mtime = np.array(model_reference_time) - np.datetime64(
            "1970-01-01T00:00:00Z", "s"
        )
        mtime = mtime.astype("timedelta64[s]").astype("int")

        ds = xr.Dataset(
            data_vars={
                "fcst": (["time", "leadtime", "location"], fcst),
                "obs": (["time", "leadtime", "location"], obs),
                "lon": (["location"], ds_obs.lon.data),
                "lat": (["location"], ds_obs.lat.data),
                "altitude": (["location"], ds_obs.altitude.data),
            },
            coords={
                "time": (["time"], mtime, {"units": "seconds since 1970-01-01"}),
                "leadtime": (["leadtime"], leadtime, {"units": "hours"}),
                "location": location,
            },
            attrs={"long_name": self.var.name, "units": self.var.unit},
        )
        self.ds = ds

    def save_as(self, fname, fmt="netcdf", force=False):
        """Save the dataset.

        Args:
            fname (str): Filename
            fmt (str, optional): Format. Defaults to "netcdf".
            force (bool, optional): Overwrite file if existing. Defaults to False.

        Raises:
            FileExistsError: File exists
            NotImplementedError: Missing ascii implementation

        """
        if fmt == "netcdf":
            self.ds.to_netcdf(fname)
        elif fmt == "ascii":
            if os.path.exists(fname) and not force:
                raise FileExistsError(f"Output file {fname} already exists")
            with open(fname, mode="w", encoding="utf-8") as fhandler:
                fhandler.write(f"# variable: {self.var.name}\n")
                if self.var.unit is not None:
                    fhandler.write(f"# units: {self.var.unit}\n")
                fhandler.write(
                    "date   hour  leadtime   location  lat   lon"
                    + "   altitude   obs   fcst\n"
                )
                t = 0
                for frt in self.ds["time"].data:
                    ilt = 0
                    for lt in self.ds["leadtime"].data:
                        loc = 0
                        for stid in self.ds["location"]:
                            lon = self.ds["lon"][loc]
                            lat = self.ds["lat"][loc]
                            hgt = self.ds["altitude"][loc]
                            obs = self.ds["obs"][t][ilt][loc]
                            fcst = self.ds["fcst"][t][ilt][loc]
                            frt_dt = datetime.strptime(
                                frt.astype(str)[:-3], "%Y-%m-%dT%H:%M:%S.%f"
                            )
                            fhandler.write(
                                f'{frt_dt.strftime("%Y%m%d")} {frt_dt.strftime("%H")}'
                                + f" {lt} {stid} {lat} {lon} {hgt} {obs} {fcst}\n"
                            )
                            loc = loc + 1
                        ilt = ilt + 1
                    t = t + 1
        else:
            raise NotImplementedError


class VerifDataFromFile(VerifData):
    """Create VerifData from a verif file.

    Args:
        VerifData (_type_): _description_
    """

    def __init__(self, filename, fmt="netcdf"):
        """Construct Verif dataset from file.

        Args:
            filename (str): Filename
            fmt (str, optional): _description_. Defaults to "netcdf".

        Raises:
            NotImplementedError: _description_
            FileNotFoundError: _description_
            NotImplementedError: _description_

        """
        if fmt == "netcdf":
            ds = xr.open_dataset(filename, engine="netcdf4")
        elif fmt == "ascii":
            point_values = []  # noqa
            obs_values = []  # noqa
            try:
                with open(filename, mode="r", encoding="utf-8") as fhandler:
                    hdr = ""
                    mapping = {}
                    for rline in fhandler.readlines():
                        if rline.find("#") >= 0:
                            pass
                        elif hdr == "":
                            hdr = rline.split()
                            for ind, recname in enumerate(hdr):
                                mapping.update({recname: ind})
                        else:
                            line = rline.split()
                            date = line[mapping["date"]]
                            hour = int(line[mapping["hour"]])
                            lead_time = int(line[mapping["leadtime"]])
                            location = str(line[mapping["location"]])  # noqa
                            lat = float(line[mapping["lat"]])  # noqa
                            lon = float(line[mapping["lon"]])  # noqa
                            try:
                                altitude = int(float(line[mapping["altitude"]]))
                            except ValueError:
                                altitude = np.nan  # noqa
                            try:
                                obs = float(line[mapping["obs"]])
                            except ValueError:
                                obs = np.nan  # noqa
                            try:
                                fcst = float(line[mapping["fcst"]])
                            except ValueError:
                                fcst = np.nan  # noqa
                            frt = as_datetime(date + "00")
                            frt = frt + as_timedelta(hour * 3600)
                            validtime = frt + as_timedelta(lead_time * 3600)  # noqa
                ds = xr.Dataset()
                # TODO
                raise NotImplementedError("Please finish implementation!")
            except FileNotFoundError:
                raise FileNotFoundError from FileNotFoundError
        else:
            raise NotImplementedError

        var = VerifVariable(ds.attrs["long_name"], ds.attrs["units"])
        VerifData.__init__(self, ds, var)


def parse_args_converter2ds(argv):
    """Parse the command line input arguments for setting a .

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    # Same main parser as usual
    parser = ArgumentParser("converter2ds")

    # Usual arguments which are applicable for the whole script / top-level args
    parser.add_argument(
        "--force",
        dest="force",
        help="Force re-creation of output",
        action="store_true",
        required=False,
    )
    parser.add_argument(
        "-g",
        "--geo",
        dest="geo",
        type=str,
        help="Domain/points json geometry definition file",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--station-list",
        dest="stationlist",
        type=str,
        help="Station list file",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--out-variable",
        dest="out_variable",
        type=str,
        help="Output variable in data set",
        default=None,
        required=False,
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        help="Output file",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--validtime",
        dest="validtime",
        type=str,
        help="Validtime",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--obs",
        dest="are_observations",
        help="Observation set",
        action="store_true",
        default=False,
    )
    parser.add_argument("--debug", help="Show debug information", action="store_true")

    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args, __ = parser.parse_known_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return parser, kwargs


def converter2ds(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.

    Raises:
        FileExistsError: _description_
        RuntimeError: "No geo information"

    """
    if argv is None:
        argv = sys.argv[1:]
    parser, kwargs = parse_args_converter2ds(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ converter2ds ******************")

    converter = get_multi_converters(parser, [], argv)
    try:
        validtime = kwargs["validtime"]
        if isinstance(validtime, str):
            validtime = as_datetime(kwargs["validtime"])
    except KeyError:
        validtime = None
    try:
        basetime = kwargs["basetime"]
        if isinstance(basetime, str):
            basetime = as_datetime(kwargs["basetime"])
    except KeyError:
        basetime = None

    out_variable = kwargs["out_variable"]

    unit = kwargs.get("unit")
    try:
        output = kwargs["output"]
    except KeyError:
        output = None
    force = False
    if "force" in kwargs:
        force = kwargs["force"]

    geo = None
    stationlist = None
    if "stationlist" in kwargs and kwargs["stationlist"] is not None:
        stationlist = StationList(kwargs["stationlist"])
    elif "geo" in kwargs:
        with open(kwargs["geo"], mode="r", encoding="utf8") as fhandler:
            geo_json = json.load(fhandler)
        geo = get_geo_object(geo_json)
    else:
        raise RuntimeError("No geo information")
    cache = None

    force = True
    variable = VerifVariable(out_variable, unit=unit)
    if geo is not None:
        vdata = DataFromSurfexConverter(converter, variable, geo, validtime, cache=cache)
    elif kwargs["are_observations"]:
        vdata = ObsDataFromSurfexConverter(
            converter, variable, stationlist, validtime, cache=cache
        )
    elif not kwargs["are_observations"]:
        vdata = VerificationDataFromSurfexConverter(
            converter, variable, stationlist, validtime, basetime, cache=cache
        )
    if output is not None:
        if os.path.exists(output) and not force:
            raise FileExistsError
        vdata.ds.to_netcdf(output)


def parse_args_ds2verif(argv):
    """Parse the command line input arguments for setting a .

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    # Same main parser as usual
    parser = ArgumentParser("ds2verif")

    # Usual arguments which are applicable for the whole script / top-level args
    parser.add_argument(
        "-m",
        "--model",
        dest="model_ds",
        type=str,
        help="Model dataset",
        default=None,
        required=True,
    )
    parser.add_argument(
        "-o",
        "--obs",
        dest="obs_ds",
        type=str,
        help="Observation dataset",
        default=None,
        required=True,
    )
    parser.add_argument(
        "-v",
        "--verif",
        dest="verif_ds",
        type=str,
        help="Verif dataset",
        default=None,
        required=True,
    )
    parser.add_argument(
        "--varname",
        dest="varname",
        type=str,
        help="varname",
        default=None,
        required=True,
    )
    parser.add_argument(
        "--unit",
        dest="unit",
        type=str,
        help="Unit",
        default=None,
        required=True,
    )
    parser.add_argument("--debug", help="Show debug information", action="store_true")

    variable_parse_options(parser)

    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    return parse_args_variable(parser, {}, argv)


def ds2verif(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_ds2verif(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ ds2verif ******************")

    model_ds = kwargs["model_ds"]
    obs_ds = kwargs["obs_ds"]
    verif_ds = kwargs["verif_ds"]
    ds_model = xr.open_dataset(model_ds, engine="netcdf4")
    ds_obs = xr.open_dataset(obs_ds, engine="netcdf4")

    var = VerifVariable(kwargs["varname"], kwargs["unit"])
    vds = VerifData(xr.Dataset(), var)
    vds.merge(ds_model, ds_obs)
    vds.ds.to_netcdf(verif_ds)


def parse_args_concat_datasets(argv):
    """Parse the command line input arguments for setting a .

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    # Same main parser as usual
    parser = ArgumentParser("concat_datasets")

    # Usual arguments which are applicable for the whole script / top-level args
    parser.add_argument(
        dest="datasets", type=str, nargs="*", help="Datasets", default=None
    )
    parser.add_argument(
        "-o",
        dest="output",
        type=str,
        help="Observation dataset",
        default=None,
        required=True,
    )
    parser.add_argument("--debug", help="Show debug information", action="store_true")
    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def concat_datasets(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_concat_datasets(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ concat_datasets ******************")

    output = kwargs["output"]
    datasets = kwargs["datasets"]
    dsets = []
    for dset in datasets:
        dsets.append(xr.open_dataset(dset, engine="netcdf4"))

    ds = xr.merge(dsets, compat="override")
    ds.to_netcdf(output)


def concat_datasets_verif(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_concat_datasets(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ concat_datasets_verif ******************")

    output = kwargs["output"]
    datasets = kwargs["datasets"]
    dsets = []
    for dset in datasets:
        dsets.append(
            xr.open_dataarray(
                dset, drop_variables=["longitude", "latitude"], engine="netcdf4"
            )
        )

    ds = xr.concat(dsets, dim="time")
    ds.to_netcdf(output)


def parse_args_vfld2ds(argv):
    """Parse the command line input arguments for setting a .

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    # Same main parser as usual
    parser = ArgumentParser("vfld2ds")

    # Usual arguments which are applicable for the whole script / top-level args
    parser.add_argument(
        "--inputfile",
        dest="vfldfile",
        type=str,
        help="Datasets",
        default=None,
        required=True,
    )
    parser.add_argument(
        "--variable",
        dest="variable",
        type=str,
        help="Datasets",
        default=None,
        required=True,
    )
    parser.add_argument(
        "--basetime", dest="basetime", type=str, help="Datasets", default=None
    )
    parser.add_argument(
        "--validtime",
        dest="validtime",
        type=str,
        help="Datasets",
        default=None,
        required=True,
    )
    parser.add_argument(
        "-o",
        dest="output",
        type=str,
        help="Dataset file",
        default=None,
        required=True,
    )
    parser.add_argument("--debug", help="Show debug information", action="store_true")
    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def vfld2ds(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_vfld2ds(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ vfld2ds ******************")

    output = kwargs["output"]
    vfldfile = kwargs["vfldfile"]
    basetime = as_datetime(kwargs["basetime"])
    validtime = as_datetime(kwargs["validtime"])
    try:
        variable = kwargs["variable"]
    except KeyError:
        variable = None
    vfld = DatasetFromVfld(vfldfile, basetime, validtime, variable, stationlist=None)
    vfld.ds.to_netcdf(output)


def parse_args_converter2harp(argv):
    """Get args for converter2harp.

    Parse the command line input arguments for creating SQLite
    tables for HARP from a converter.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    # Same main parser as usual
    parser = ArgumentParser("converter2harp")

    # Usual arguments which are applicable for the whole script / top-level args
    parser.add_argument(
        "--station-list",
        dest="stationlist",
        type=str,
        help="Domain/points json geometry definition file",
        default=None,
        required=False,
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        help="Output SQLite template",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--validtime",
        dest="validtime",
        type=str,
        help="Validtime",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--harp-param",
        dest="harp-param",
        type=str,
        help="HARP parameter name",
        required=True,
    )
    parser.add_argument(
        "--harp-param-unit",
        dest="harp-param-unit",
        type=str,
        help="HARP parameter unit",
        required=True,
    )
    parser.add_argument(
        "--model-name",
        dest="model-name",
        type=str,
        help="HARP model name",
        required=True,
    )
    parser.add_argument("--debug", help="Show debug information", action="store_true")

    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args, __ = parser.parse_known_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})

    return parser, kwargs


def converter2harp_cli(argv=None):
    """Create harp data froma converter.

    Args:
        argv (list, optional): Argument list. Defaults to None

    """
    if argv is None:
        argv = sys.argv[1:]
    parser, kwargs = parse_args_converter2harp(argv)

    converter = get_multi_converters(parser, [], argv)
    converter2harp(converter, **kwargs)


def converter2harp(converter, **kwargs):
    """Create harp data froma converter.

    Args:
        converter (Converter): Converter object
        kwargs (dict): Key-word arguments

    """
    try:
        validtime = kwargs["validtime"]
        if isinstance(validtime, str):
            validtime = as_datetime(kwargs["validtime"])
    except KeyError:
        validtime = None
    try:
        harp_param_level = kwargs["harp-param-level"]
    except KeyError:
        harp_param_level = None
    try:
        harp_param_level_name = kwargs["harp-param-level-name"]
    except KeyError:
        harp_param_level_name = None

    harp_param = kwargs["harp-param"]
    harp_param_unit = kwargs["harp-param-unit"]
    model_name = kwargs["model-name"]

    basetime = converter.initial_time
    variable = converter.name
    try:
        sqlite_template = kwargs["output"]
    except KeyError:
        sqlite_template = "{MODEL}/{YYYY}/{MM}/FCTABLE_{PP}_{YYYY}{MM}.sqlite"

    try:
        stationlist = StationList(kwargs["stationlist"])
    except KeyError:
        raise RuntimeError from KeyError

    geo = stationlist.geo
    cache = None

    if sqlite_template is None:
        sqlite_template = "{MODEL}/{YYYY}/{MM}/FCTABLE_{PP}_{YYYY}{MM}.sqlite"

    leadtime = int((validtime - basetime).total_seconds() / 3600.0)
    cache = None
    param = {
        "varname": variable,
        "harp_param": harp_param,
        "units": harp_param_unit,
        "level": harp_param_level,
        "level_name": harp_param_level_name,
    }

    data_vector = ConvertedInput(geo, variable, converter).read_time_step(
        validtime, cache
    )

    lons, lats, elevs = stationlist.get_pos_from_stid(stationlist.stids)
    harp_station_list = pandas.DataFrame(
        {
            "SID": stationlist.ids,
            "lat": lats,
            "lon": lons,
            "elev": elevs,
            "name": stationlist.ids,
        }
    )

    logging.info("SQLITE: writing var=%s for mode=%sn", variable, model_name)
    sqlite_file = sqlite_name(param, validtime, model_name, sqlite_template)
    data = create_table(
        data_vector, harp_station_list, param, validtime, leadtime, model_name + "_det"
    )
    logging.info("SQLITE: writing sqlite_file=%s", sqlite_file)
    write_to_sqlite(data, sqlite_file, param, model_name + "_det")



"""Command line interfaces."""

import contextlib
import json
import logging
import os
import sys

import numpy as np
import yaml

try:
    import matplotlib.pyplot as plt
except ModuleNotFoundError:
    plt = None


from .binary_input import JsonOutputDataFromFile
from .cache import Cache
from .cmd_parsing import (
    parse_args_bufr2json,
    parse_args_create_forcing,
    parse_args_create_namelist,
    parse_args_dump_environ,
    parse_args_first_guess_for_oi,
    parse_args_gridpp,
    parse_args_lsm_file_assim,
    parse_args_merge_qc_data,
    parse_args_modify_forcing,
    parse_args_obs2json,
    parse_args_oi2soda,
    parse_args_plot_points,
    parse_args_qc2obsmon,
    parse_args_set_geo_from_obs_set,
    parse_args_set_geo_from_stationlist,
    parse_args_shape2ign,
    parse_args_surfex_binary,
    parse_args_titan,
    parse_cryoclim_pseudoobs,
    parse_sentinel_obs,
    parse_set_domain,
)
from .datetime_utils import as_datetime, as_datetime_args, as_timedelta, get_decade
from .file import PGDFile, PREPFile, SURFFile
from .forcing import modify_forcing, run_time_loop, set_forcing_config
from .geo import ConfProjFromHarmonie, LonLatVal, get_geo_object, shape2ign
from .input_methods import create_obsset_file, get_datasources, set_geo_from_obs_set
from .interpolation import horizontal_oi
from .namelist import NamelistGeneratorAssemble, NamelistGeneratorFromNamelistFile
from .netcdf import (
    create_netcdf_first_guess_template,
    oi2soda,
    read_first_guess_netcdf_file,
    write_analysis_netcdf_file,
)
from .obs import StationList
from .obsmon import write_obsmon_sqlite_file
from .platform_deps import SystemFilePathsFromFile
from .pseudoobs import CryoclimObservationSet, SentinelObservationSet
from .read import ConvertedInput, Converter, get_multi_converters
from .run import BatchJob, PerturbedOffline, SURFEXBinary
from .titan import (
    TitanDataSet,
    dataset_from_file,
    define_quality_control,
    merge_json_qc_data_sets,
)
from .variable import Variable


def run_surfex_binary(mode, **kwargs):
    """Run a surfex binary."""
    logging.debug("ARGS: %s", kwargs)

    try:
        masterodb = kwargs["masterodb"]
    except KeyError:
        masterodb = False

    domain = kwargs.get("domain")
    geo = None
    if domain is not None:
        with open(domain, mode="r", encoding="utf-8") as fhandler:
            geo = get_geo_object(json.load(fhandler))

    system_file_paths = kwargs["system_file_paths"]
    if os.path.exists(system_file_paths):
        system_file_paths = SystemFilePathsFromFile(system_file_paths)
    else:
        raise FileNotFoundError("File not found: " + system_file_paths)

    if "forcing_dir" in kwargs:
        system_file_paths.add_system_file_path("forcing_dir", kwargs["forcing_dir"])

    do_pgd = False
    do_prep = False
    do_perturbed = False
    need_pgd = True
    need_prep = True
    try:
        one_decade = kwargs["one_decade"]
    except KeyError:
        one_decade = False
    check_existence = True
    if kwargs.get("tolerate_missing"):
        check_existence = False

    try:
        nml_macros = kwargs["macros"]
        if isinstance(nml_macros, str):
            with open(nml_macros, mode="r", encoding="utf8") as fhandler:
                nml_macros = json.load(fhandler)
    except KeyError:
        nml_macros = {}

    prep_input_file = None
    if "prep_file" in kwargs:
        prep_input_file = kwargs["prep_file"]
    prep_input_pgdfile = None
    if "prep_pgdfile" in kwargs:
        prep_input_pgdfile = kwargs["prep_pgdfile"]

    nml_macros.update(
        {
            "PREP_INPUT_FILE_WITH_PATH": prep_input_file,
            "PREP_PGD_INPUT_FILE_WITH_PATH": prep_input_pgdfile,
        }
    )

    # TODO
    perturbed_file_pattern = None
    if perturbed_file_pattern in kwargs:
        perturbed_file_pattern = kwargs["perturbed_file_pattern"]

    basetime = None
    if "basetime" in kwargs and kwargs["basetime"] is not None:
        if isinstance(kwargs["basetime"], str):
            basetime = as_datetime(kwargs["basetime"])
            kwargs.update({"basetime": basetime})
        else:
            basetime = kwargs["basetime"]
    try:
        validtime = kwargs["validtime"]
    except KeyError:
        validtime = basetime
    if basetime is not None:
        xtime = (
            basetime - basetime.replace(hour=0, second=0, microsecond=0)
        ).total_seconds()
        nml_macros.update(
            {
                "PREP_NYEAR": basetime.year,
                "PREP_NMONTH": basetime.month,
                "PREP_NDAY": basetime.day,
                "PREP_XTIME": xtime,
                "SODA_HH": f"{basetime.hour:02d}",
            }
        )
        if one_decade:
            nml_macros.update({"DECADE": get_decade(basetime)})
    else:
        if one_decade:
            raise RuntimeError("You must set basetime with option one_decade")
        if mode in ("prep", "soda"):
            raise RuntimeError(f"You must set basetime when using mode {mode}")

    logging.debug("kwargs: %s", str(kwargs))
    binary = kwargs["binary"]
    rte = kwargs["rte"]
    wrapper = kwargs["wrapper"]
    namelist_path = kwargs["namelist_path"]
    force = kwargs["force"]
    output = kwargs["output"]
    masterodb = kwargs["masterodb"]
    archive = kwargs["archive"]
    only_archive = kwargs.get("only_archive", False)
    print_namelist = kwargs["print_namelist"]
    logging.debug("masterodb %s", masterodb)

    forc_zs = kwargs.get("forc_zs", False)
    try:
        assemble_file = kwargs["assemble_file"]
    except KeyError:
        assemble_file = None

    if mode == "pgd":
        do_pgd = True
        need_pgd = False
        need_prep = False
    elif mode == "prep":
        do_prep = True
        need_prep = False
    elif mode in ("offline", "soda"):
        pass
    elif mode == "perturbed":
        do_perturbed = True
        mode = "offline"
    else:
        raise NotImplementedError(mode + " is not implemented!")

    pgd_file_path = None
    if need_pgd:
        pgd_file_path = kwargs["pgd"]

    prep_file_path = None
    if need_prep:
        prep_file_path = kwargs["prep"]

    pert = None
    if "pert" in kwargs:
        pert = kwargs["pert"]
    negpert = False
    if "negpert" in kwargs:
        negpert = kwargs["negpert"]

    try:
        basetime = kwargs["basetime"]
        if isinstance(basetime, str):
            basetime = as_datetime(basetime)
    except KeyError:
        basetime = None

    try:
        input_binary_data = kwargs["input_binary_data"]
    except KeyError:
        input_binary_data = None

    # Create run time enviroment from local unless given
    if rte is None:
        rte = "rte.json"
    if not os.path.exists(rte):
        with open(rte, mode="w", encoding="utf-8") as fhandler:
            json.dump(os.environ.copy(), fhandler)

    with open(rte, mode="r", encoding="utf-8") as file_handler:
        rte = json.load(file_handler)
        my_batch = BatchJob(rte, wrapper=wrapper)

    my_archive = None
    if archive is not None:
        if os.path.exists(archive):
            my_archive = JsonOutputDataFromFile(archive)
        else:
            raise FileNotFoundError("File not found: " + archive)

    logging.info("nml_macros: %s", nml_macros)
    if not (output is not None and os.path.exists(output)) or force:
        if assemble_file is None:
            nam_gen = NamelistGeneratorFromNamelistFile(
                mode, namelist_path, macros=nml_macros
            )
        else:
            with open(namelist_path, mode="r", encoding="utf-8") as file_handler:
                nam_defs = yaml.safe_load(file_handler)
            with open(assemble_file, mode="r", encoding="utf8") as file_handler:
                assemble = yaml.safe_load(file_handler)
            nam_gen = NamelistGeneratorAssemble(
                mode, nam_defs, assemble, macros=nml_macros
            )

        if mode == "pgd":
            nam_gen = geo.update_namelist(nam_gen)
        my_settings = nam_gen.get_namelist()
        if mode == "pgd" and one_decade:
            try:
                ntime = my_settings["nam_data_isba"]["ntime"]
                if ntime != 1:
                    logging.warning(
                        "Overriding value %s nam_data_isba#ntime and setting it 1",
                        "because of one_decade=True",
                    )
                    my_settings["nam_data_isba"]["ntime"] = 1
            except KeyError:
                raise RuntimeError from KeyError
        if forc_zs:
            try:
                val = my_settings["nam_io_offline"]["lset_forc_zs"]
            except KeyError:
                val = False
            if not val:
                logging.warning("Override lset_forc_zs with value True")
            my_settings["nam_io_offline"]["lset_forc_zs"] = True
        with contextlib.suppress(KeyError):
            my_settings["nam_io_offline"]["xtstep_output"] = kwargs["output_frequency"]

        with contextlib.suppress(KeyError):
            my_settings["nam_var"]["nvar"] = sum(my_settings["nam_var"]["nncv"])

        try:
            nnco = my_settings["nam_obs"]["nnco"]
        except KeyError:
            nnco = None
        if nnco is not None:
            nobstype = nnco if isinstance(nnco, int) else int(sum(nnco))
            try:
                nobstype_old = my_settings["nam_obs"]["nobstype"]
            except KeyError:
                nobstype_old = None
            if nobstype_old is not None and nobstype_old != nobstype:
                logging.warning(
                    "Mismatch in nobstype. Override %s with: %s", nobstype_old, nobstype
                )
            my_settings["nam_obs"]["nobstype"] = nobstype

        if input_binary_data is None:
            raise RuntimeError("input_binary_data not set")
        with open(input_binary_data, mode="r", encoding="utf-8") as file_handler:
            input_binary_data = json.load(file_handler)

        input_data = nam_gen.input_data_from_namelist(
            input_binary_data,
            system_file_paths,
            validtime=validtime,
            basetime=basetime,
            check_existence=check_existence,
        )

        # Create input
        my_format = my_settings["nam_io_offline"]["csurf_filetype"]
        my_pgdfile = my_settings["nam_io_offline"]["cpgdfile"]
        my_prepfile = my_settings["nam_io_offline"]["cprepfile"]
        my_surffile = my_settings["nam_io_offline"]["csurffile"]
        lfagmap = False
        if "lfagmap" in my_settings["nam_io_offline"]:
            lfagmap = my_settings["nam_io_offline"]["lfagmap"]

        logging.debug("pgdfile=%s lfagmap=%s %s", my_pgdfile, lfagmap, pgd_file_path)
        logging.debug("nam_io_offline=%s", my_settings["nam_io_offline"])

        if not only_archive:
            if need_pgd:
                logging.debug("Need pgd")
                my_pgdfile = PGDFile(
                    my_format,
                    my_pgdfile,
                    input_file=pgd_file_path,
                    lfagmap=lfagmap,
                    masterodb=masterodb,
                )

            if need_prep:
                logging.debug("Need prep")
                my_prepfile = PREPFile(
                    my_format,
                    my_prepfile,
                    input_file=prep_file_path,
                    lfagmap=lfagmap,
                    masterodb=masterodb,
                )

            surffile = None
            if need_prep and need_pgd:
                logging.debug("Need pgd and prep")
                surffile = SURFFile(
                    my_format,
                    my_surffile,
                    archive_file=output,
                    lfagmap=lfagmap,
                    masterodb=masterodb,
                )

            if do_perturbed:
                PerturbedOffline(
                    binary,
                    my_batch,
                    my_prepfile,
                    pert,
                    my_settings,
                    input_data,
                    pgdfile=my_pgdfile,
                    surfout=surffile,
                    archive_data=my_archive,
                    print_namelist=print_namelist,
                    negpert=negpert,
                )
            elif do_pgd:
                my_pgdfile = PGDFile(
                    my_format,
                    my_pgdfile,
                    input_file=pgd_file_path,
                    archive_file=output,
                    lfagmap=lfagmap,
                    masterodb=masterodb,
                )
                SURFEXBinary(
                    binary,
                    my_batch,
                    my_pgdfile,
                    my_settings,
                    input_data,
                    archive_data=my_archive,
                    print_namelist=print_namelist,
                )
            elif do_prep:
                my_prepfile = PREPFile(
                    my_format,
                    my_prepfile,
                    archive_file=output,
                    lfagmap=lfagmap,
                    masterodb=masterodb,
                )
                SURFEXBinary(
                    binary,
                    my_batch,
                    my_prepfile,
                    my_settings,
                    input_data,
                    pgdfile=my_pgdfile,
                    archive_data=my_archive,
                    print_namelist=print_namelist,
                )
            else:
                SURFEXBinary(
                    binary,
                    my_batch,
                    my_prepfile,
                    my_settings,
                    input_data,
                    pgdfile=my_pgdfile,
                    surfout=surffile,
                    archive_data=my_archive,
                    print_namelist=print_namelist,
                )
    else:
        logging.info("%s already exists!", output)


def run_gridpp(**kwargs):
    """Gridpp."""
    var = kwargs["var"]
    input_file = kwargs["input_file"]
    output_file = None
    if "output_file" in kwargs:
        output_file = kwargs["output_file"]
    hlength = kwargs["hlength"]
    vlength = 100000
    if "vlength" in kwargs:
        vlength = kwargs["vlength"]
    wlength = 0
    if "wlength" in kwargs:
        wlength = kwargs["wlength"]
    max_locations = 20
    if "max_locations" in kwargs:
        max_locations = kwargs["max_locations"]
    elev_gradient = 0
    if "elev_gradient" in kwargs:
        elev_gradient = kwargs["elev_gradient"]
    if elev_gradient not in (-0.0065, 0):
        raise RuntimeError("Not a valid elevation gradient")
    epsilon = None
    if "epsilon" in kwargs:
        epsilon = kwargs["epsilon"]
    minvalue = None
    if "minvalue" in kwargs:
        minvalue = kwargs["minvalue"]
    maxvalue = None
    if "maxvalue" in kwargs:
        maxvalue = kwargs["maxvalue"]
    only_diff = False
    if "only_diff" in kwargs:
        only_diff = kwargs["only_diff"]
    obs_file = kwargs["obs_file"]

    # Get input fields
    geo, validtime, background, glafs, gelevs = read_first_guess_netcdf_file(
        input_file, var
    )

    an_time = validtime
    # Read OK observations
    observations = dataset_from_file(an_time, obs_file, qc_flag=0)
    logging.info("Found %s observations with QC flag == 0", str(len(observations.lons)))
    field = horizontal_oi(
        geo,
        background,
        observations,
        gelevs,
        hlength=hlength,
        vlength=vlength,
        wlength=wlength,
        structure_function="Barnes",
        max_locations=max_locations,
        elev_gradient=elev_gradient,
        epsilon=epsilon,
        minvalue=minvalue,
        maxvalue=maxvalue,
        interpol="bilinear",
        only_diff=only_diff,
    )

    if output_file is not None:
        write_analysis_netcdf_file(
            output_file, field, var, validtime, gelevs, glafs, new_file=True, geo=geo
        )


def run_titan(**kwargs):
    """Titan."""
    domain = kwargs.get("domain")
    domain_geo = None
    if domain is not None:
        with open(domain, mode="r", encoding="utf-8") as fhandler:
            domain_geo = get_geo_object(json.load(fhandler))

    blacklist = None
    if "blacklist" in kwargs:
        blacklist = kwargs["blacklist"]
    elif "blacklist_file" in kwargs and kwargs["blacklist_file"] is not None:
        with open(kwargs["blacklist_file"], mode="r", encoding="utf-8") as fhandler:
            blacklist = json.load(fhandler)

    if "input_file" in kwargs:
        input_file = kwargs["input_file"]
        if os.path.exists(input_file):
            with open(input_file, "r", encoding="utf-8") as fhandler:
                settings = json.load(fhandler)
        else:
            raise FileNotFoundError("Could not find input file " + input_file)
    elif "input_data" in kwargs:
        settings = kwargs["input_data"]
    else:
        raise RuntimeError("You must specify input_file or input_data")

    tests = kwargs["tests"]
    output_file = None
    if "output_file" in kwargs:
        output_file = kwargs["output_file"]
    indent = None
    if "indent" in kwargs:
        indent = kwargs["indent"]

    an_time = kwargs["validtime"]
    if isinstance(an_time, str):
        an_time = as_datetime(an_time)
    var = kwargs["variable"]

    tests = define_quality_control(
        tests, settings[var], an_time, domain_geo=domain_geo, blacklist=blacklist
    )
    logging.debug("Settings: %s", settings)
    datasources = get_datasources(an_time, settings[var]["sets"])
    data_set = TitanDataSet(var, settings[var], tests, datasources, an_time)
    data_set.perform_tests()

    if output_file is not None:
        data_set.write_output(output_file, indent=indent)


def run_oi2soda(**kwargs):
    """Oi2soda."""
    t2m_file = kwargs["t2m_file"]
    rh2m_file = kwargs["rh2m_file"]
    sd_file = kwargs["sd_file"]
    sm_file = kwargs["sm_file"]
    output = kwargs["output"]

    t2m = None
    if t2m_file is not None:
        t2m = {"file": t2m_file, "var": kwargs["t2m_var"]}
    rh2m = None
    if rh2m_file is not None:
        rh2m = {"file": rh2m_file, "var": kwargs["rh2m_var"]}
    s_d = None
    if sd_file is not None:
        s_d = {"file": sd_file, "var": kwargs["sd_var"]}
    s_m = None
    if sm_file is not None:
        s_m = {"file": sm_file, "var": kwargs["sm_var"]}

    basetime = as_datetime(kwargs["basetime"])
    oi2soda(basetime, t2m=t2m, rh2m=rh2m, s_d=s_d, s_m=s_m, output=output)


def set_geo_from_stationlist(**kwargs):
    """Set geometry from station list."""
    stationlist = kwargs["stationlist"]
    lonrange = None
    if "lonrange" in kwargs:
        lonrange = kwargs["lonrange"]
    latrange = None
    if "latrange" in kwargs:
        latrange = kwargs["latrange"]
    if lonrange is None:
        lonrange = [-180, 180]
    if latrange is None:
        latrange = [-90, 90]

    lons = []
    lats = []
    stationlist = StationList(stationlist)
    stids = stationlist.stids

    for stid in stids:
        lon, lat, __ = stationlist.get_pos_from_stid([stid])
        lon = lon[0]
        lat = lat[0]
        if lonrange[0] <= lon <= lonrange[1] and latrange[0] <= lat <= latrange[1]:
            lon = round(lon, 5)
            lat = round(lat, 5)
            lons.append(lon)
            lats.append(lat)

    d_x = ["0.3"] * len(lons)
    geo_json = {
        "nam_pgd_grid": {"cgrid": "LONLATVAL"},
        "nam_lonlatval": {"xx": lons, "xy": lats, "xdx": d_x, "xdy": d_x},
    }
    return LonLatVal(geo_json)


def sentinel_obs(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_sentinel_obs(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ sentinel_obs ******************")
    fg_file = kwargs["fg_file"]
    infiles = kwargs["infiles"]
    step = kwargs["thinning"]
    output = kwargs["output"]
    varname = kwargs["varname"]
    indent = kwargs["indent"]

    fg_geo, validtime, grid_sm_fg, __, __ = read_first_guess_netcdf_file(fg_file, varname)
    obs_set = SentinelObservationSet(infiles, validtime, fg_geo, grid_sm_fg, step=step)
    obs_set.write_json_file(output, indent=indent)


def qc2obsmon(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_qc2obsmon(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ qc2obsmon ******************")
    write_obsmon_sqlite_file(**kwargs)


def prep(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_surfex_binary(argv, "prep")
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ prep ******************")
    run_surfex_binary("prep", **kwargs)


def plot_points(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.

    Raises:
        NotImplementedError: "Inputype is not implemented"
        RuntimeError: "No geo is set"
        RuntimeError: "No field read"
        ModuleNotFoundError: "Matplotlib is needed to plot"

    """
    if argv is None:
        argv = sys.argv[1:]
    parser, kwargs = parse_args_plot_points(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ plot_points ******************")

    validtime = as_datetime(kwargs["validtime"])
    output = kwargs["output"]

    geo_file = None
    if "geo" in kwargs:
        geo_file = kwargs["geo"]

    geo = None
    if geo_file is not None:
        with open(geo_file, mode="r", encoding="utf-8") as fhandler:
            domain_json = json.load(fhandler)
        geo = get_geo_object(domain_json)

    contour = True
    if "no_contour" in kwargs:
        no_contour = kwargs["no_contour"]
        if no_contour:
            contour = False

    cache = Cache(-1)
    converter = get_multi_converters(parser, [], argv)
    field = ConvertedInput(geo, "var", converter).read_time_step(validtime, cache)

    if field is None:
        raise RuntimeError("No field read")

    converter_name = converter.name
    try:
        title = kwargs["title"]
    except KeyError:
        title = None
    if title is None:
        title = f"converter={converter_name}"

    if geo is None:
        raise RuntimeError("No geo is set")

    if field is None:
        raise RuntimeError("No field read")

    logging.debug(
        "npoints=%s nlons=%s nlats=%s contour=%s field.shape=%s",
        geo.npoints,
        geo.nlons,
        geo.nlats,
        contour,
        field.shape,
    )
    if geo.npoints not in (geo.nlons, geo.nlats):
        if contour:
            field = np.reshape(field, [geo.nlons, geo.nlats])
    else:
        contour = False

    if plt is None:
        raise ModuleNotFoundError("Matplotlib is needed to plot")
    logging.debug(
        "lons.shape=%s lats.shape=%s field.shape=%s",
        geo.lons.shape,
        geo.lats.shape,
        field.shape,
    )
    if contour:
        plt.contourf(geo.lons, geo.lats, field)
    else:
        plt.scatter(geo.lonlist, geo.latlist, c=field)

    plt.title(title)
    plt.colorbar()
    if output is None:
        plt.show()
    else:
        logging.info("Saving figure in %s", output)
        plt.savefig(output)


def pgd(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_surfex_binary(argv, "pgd")
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ pgd ******************")
    run_surfex_binary("pgd", **kwargs)


def perturbed_offline(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_surfex_binary(argv, "perturbed")
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ perturbed ******************")
    run_surfex_binary("perturbed", **kwargs)


def offline(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_surfex_binary(argv, "offline")
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ offline ******************")
    run_surfex_binary("offline", **kwargs)


def cli_oi2soda(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_oi2soda(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ oi2soda ******************")
    run_oi2soda(**kwargs)


def cli_modify_forcing(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_modify_forcing(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ modify_forcing ******************")
    modify_forcing(**kwargs)


def cli_merge_qc_data(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_merge_qc_data(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ merge_qc_data ******************")

    qc_data = merge_json_qc_data_sets(kwargs.get("validtime"), kwargs.get("filenames"))
    qc_data.write_output(kwargs.get("output"), indent=kwargs.get("indent"))


def gridpp(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_gridpp(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ gridpp ******************")
    run_gridpp(**kwargs)


def dump_environ(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_dump_environ(argv)
    outputfile = kwargs["outputfile"]
    with open(outputfile, mode="w", encoding="utf-8") as file_handler:
        json.dump(os.environ.copy(), file_handler)


def first_guess_for_oi(argv=None):
    """First guess for OI.

    Args:
        argv (list): Arguments

    """
    if argv is None:
        argv = sys.argv[1:]

    parser, kwargs = parse_args_first_guess_for_oi(argv)
    debug = kwargs.get("debug")
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )

    domain = kwargs["domain"]
    with open(domain, mode="r", encoding="utf-8") as fhandler:
        geo = get_geo_object(json.load(fhandler))

    validtime = kwargs["validtime"]
    output = kwargs["output"]
    fg_variables = kwargs["fg_variables"]

    fvariables = []
    for fg_var in fg_variables:
        try:
            outvar = kwargs[f"{fg_var}_outfile_var"]
        except KeyError:
            logging.warning("Set output var equal to fg_var %s", fg_var)
            outvar = fg_var
        fvariables.append(outvar)
    validtime = as_datetime(validtime)

    converters = get_multi_converters(parser, fg_variables, argv)
    cache = Cache(3600)

    f_g = None
    for fg_var in fg_variables:
        var = kwargs[f"{fg_var}_outfile_var"]
        converter = converters[fg_var]

        field = ConvertedInput(geo, fg_var, converter).read_time_step(validtime, cache)
        field = np.reshape(field, [geo.nlons, geo.nlats])

        # Create file
        if f_g is None:
            n_x = geo.nlons
            n_y = geo.nlats
            f_g = create_netcdf_first_guess_template(
                fvariables, n_x, n_y, output, geo=geo
            )
            epoch = float(
                (validtime - as_datetime_args(year=1970, month=1, day=1)).total_seconds()
            )
            f_g.variables["time"][:] = epoch
            f_g.variables["longitude"][:] = np.transpose(geo.lons)
            f_g.variables["latitude"][:] = np.transpose(geo.lats)
            f_g.variables["x"][:] = list(range(n_x))
            f_g.variables["y"][:] = list(range(n_y))

        if var == "altitude":
            field[field < 0] = 0

        if np.isnan(np.sum(field)):
            fill_nan_value = f_g.variables[var].getncattr("_FillValue")
            logging.info("Field %s got Nan. Fill with: %s", var, str(fill_nan_value))
            field[np.where(np.isnan(field))] = fill_nan_value

        f_g.variables[var][:] = np.transpose(field)

    if f_g is not None:
        f_g.close()


def cryoclim_pseudoobs(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_cryoclim_pseudoobs(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ cryoclim_pseudoobs ******************")

    fg_file = kwargs["fg"]["filepattern"]
    varname = kwargs["fg"]["varname"]
    infiles = kwargs["infiles"]
    step = kwargs["thinning"]
    output = kwargs["output"]
    validtime = as_datetime(kwargs["validtime"], offset=True)
    indent = kwargs["indent"]
    laf_threshold = kwargs["laf_threshold"]
    cryo_varname = kwargs["cryo_varname"]
    fg_geo, validtime, grid_snow_fg, glafs, gelevs = read_first_guess_netcdf_file(
        fg_file, varname
    )

    grid_perm_snow = None
    grid_perm_snow_geo = None
    if "perm_snow" in kwargs and kwargs["perm_snow"]["filepattern"] is not None:
        fileformat = kwargs["perm_snow"]["inputtype"]
        basetime = as_datetime(kwargs["perm_snow"]["basetime"], offset=True)
        if fileformat is not None:
            grid_perm_snow, grid_perm_snow_geo = Variable(
                fileformat, kwargs["perm_snow"], basetime
            ).read_var_field(validtime)

    grid_slope = None
    grid_slope_geo = None
    if "slope" in kwargs and kwargs["slope"]["filepattern"] is not None:
        fileformat = kwargs["slope"]["inputtype"]
        basetime = as_datetime(kwargs["slope"]["basetime"], offset=True)
        if fileformat is not None:
            grid_slope, grid_slope_geo = Variable(
                fileformat, kwargs["slope"], basetime
            ).read_var_field(validtime)

    obs_set = CryoclimObservationSet(
        infiles,
        validtime,
        fg_geo,
        grid_snow_fg,
        gelevs,
        perm_snow=grid_perm_snow,
        perm_snow_geo=grid_perm_snow_geo,
        slope=grid_slope,
        slope_geo=grid_slope_geo,
        step=step,
        glaf=glafs,
        laf_threshold=laf_threshold,
        cryo_varname=cryo_varname,
    )
    obs_set.write_json_file(output, indent=indent)


def create_namelist(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.

    Raises:
        FileNotFoundError: "File not found:"
        RuntimeError: "Geo is needed for PGD"

    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_create_namelist(argv)
    debug = kwargs.get("debug")
    mode = kwargs.get("mode")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ %s ******************", mode)

    logging.debug("ARGS: %s", kwargs)
    mode = kwargs.get("mode")

    domain = kwargs.get("domain", None)
    if domain is not None:
        with open(domain, mode="r", encoding="utf-8") as fhandler:
            geo = get_geo_object(json.load(fhandler))
    if mode == "pgd" and geo is None:
        raise RuntimeError("Geo is needed for PGD")
    uppercase = kwargs.get("uppercase")
    true_repr = kwargs.get("true_repr")
    false_repr = kwargs.get("false_repr")
    if uppercase:
        false_repr = false_repr.upper()
        true_repr = true_repr.upper()
    else:
        false_repr = false_repr.lower()
        true_repr = true_repr.lower()
    namelist_defs_file = kwargs.get("namelist_defs")
    assemble_file = kwargs["assemble_file"]
    output = kwargs.get("output")
    if output is None:
        output = "OPTIONS.nam"

    if assemble_file is None:
        nam_gen = NamelistGeneratorFromNamelistFile(mode, namelist_defs_file)
    else:
        with open(namelist_defs_file, mode="r", encoding="utf-8") as file_handler:
            nam_defs = yaml.safe_load(file_handler)
        with open(assemble_file, mode="r", encoding="utf-8") as file_handler:
            assemble = yaml.safe_load(file_handler)
        nam_gen = NamelistGeneratorAssemble(mode, nam_defs, assemble)

    if geo is not None:
        nam_gen = geo.update_namelist(nam_gen)
    if os.path.exists(output):
        os.remove(output)
    nam_gen.write(output, uppercase=uppercase, true_repr=true_repr, false_repr=false_repr)


def create_lsm_file_assim(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.

    Raises:
        FileNotFoundError: Domain file not found

    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_lsm_file_assim(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ create_lsm_fil ******************")

    domain = kwargs["domain"]
    logging.debug("domain=%s", domain)
    if os.path.exists(domain):
        with open(domain, "r", encoding="utf-8") as fhandler:
            domain_json = json.load(fhandler)
        geo = get_geo_object(domain_json)
    else:
        raise FileNotFoundError(domain)
    validtime = as_datetime(kwargs["basetime"])

    # TODO Move to a method outside cli
    cache = Cache(3600)
    inputfile = kwargs["file"]
    fileformat = kwargs["fileformat"]
    converter = kwargs["converter"]
    output = kwargs["output"]

    var = kwargs["var"]

    defs = {
        "filepattern": inputfile,
        "fileformat": fileformat,
        "filetype": "surf",
        "fcint": 10800,
        "offset": 0,
    }

    logging.debug("%s %s", var, fileformat)
    converter_conf = {"none": {"name": var, "varname": var}}

    var = "LSM"
    initial_basetime = validtime - as_timedelta(seconds=10800)
    converter = Converter(converter, initial_basetime, defs, converter_conf, fileformat)
    field = ConvertedInput(geo, var, converter).read_time_step(validtime, cache)
    field = np.reshape(field, [geo.nlons, geo.nlats])
    field = np.transpose(field)

    with open(output, mode="w", encoding="utf-8") as file_handler:
        for lat in range(geo.nlats):
            for lon in range(geo.nlons):
                file_handler.write(str(field[lat, lon]) + "\n")


def create_forcing(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_create_forcing(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ create_forcing ******************")
    options, var_objs, att_objs = set_forcing_config(**kwargs)
    run_time_loop(options, var_objs, att_objs)


def bufr2json(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_bufr2json(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ bufr2json ******************")
    logging.warning("This method is depreciated. Please use create_obsset_file")
    variables = kwargs.get("vars")
    bufrfile = kwargs.get("bufr")
    output = kwargs.get("output")
    valid_dtg = as_datetime(kwargs.get("obs_time"))
    valid_range = as_timedelta(seconds=kwargs.get("valid_range"))
    sigmao = kwargs.get("sigmao")
    label = kwargs.get("label")
    indent = kwargs.get("indent")
    lonrange = kwargs.get("lonrange")
    latrange = kwargs.get("latrange")

    create_obsset_file(
        valid_dtg,
        "bufr",
        variables,
        bufrfile,
        output,
        pos_t_range=valid_range,
        lonrange=lonrange,
        latrange=latrange,
        label=label,
        indent=indent,
        sigmao=sigmao,
    )


def obs2json(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_obs2json(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ obs2json ******************")
    obs_type = kwargs.get("obs_type")
    variables = kwargs.get("vars")
    inputfile = kwargs.get("inputfile")
    output = kwargs.get("output")
    obs_time = as_datetime(kwargs.get("obs_time"))
    label = kwargs.get("label")
    unit = kwargs.get("unit")
    level = kwargs.get("level")
    obtypes = kwargs.get("obtypes")
    subtypes = kwargs.get("subtypes")
    sigmao = kwargs.get("sigmao")
    pos_t_range = kwargs.get("pos_t_range")
    neg_t_range = kwargs.get("neg_t_range")
    indent = kwargs.get("indent")
    lonrange = kwargs.get("lonrange")
    latrange = kwargs.get("latrange")
    if pos_t_range is not None:
        pos_t_range = as_timedelta(seconds=pos_t_range)
    if neg_t_range is not None:
        neg_t_range = as_timedelta(seconds=neg_t_range)

    create_obsset_file(
        obs_time,
        obs_type,
        variables,
        inputfile,
        output,
        pos_t_range=pos_t_range,
        neg_t_range=neg_t_range,
        lonrange=lonrange,
        latrange=latrange,
        label=label,
        unit=unit,
        level=level,
        indent=indent,
        obtypes=obtypes,
        subtypes=subtypes,
        sigmao=sigmao,
    )


def cli_set_domain_from_harmonie(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.

    Raises:
        FileNotFoundError: File not found
        RuntimeError: Domain not provided

    """
    if argv is None:
        argv = sys.argv[1:]

    args = parse_set_domain(argv)
    debug = args.debug

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ set_domain_from_harmonie ******************")

    output = args.output
    indent = args.indent
    geo = ConfProjFromHarmonie()
    domain_json = geo.json
    with open(output, mode="w", encoding="utf-8") as file_handler:
        json.dump(domain_json, file_handler, indent=indent)


def cli_set_geo_from_obs_set(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]

    kwargs = parse_args_set_geo_from_obs_set(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ set_geo_from_obs_set ******************")

    validtime = as_datetime(kwargs["validtime"])
    geo = set_geo_from_obs_set(
        validtime,
        kwargs["obs_type"],
        kwargs["variable"],
        kwargs["inputfile"],
        kwargs["lonrange"],
        kwargs["latrange"],
    )
    output = kwargs["output"]
    with open(output, mode="w", encoding="utf-8") as file_handler:
        json.dump(geo.json, file_handler)


def cli_set_geo_from_stationlist(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]

    kwargs = parse_args_set_geo_from_stationlist(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ set_geo_from_stationlist ******************")
    geo = set_geo_from_stationlist(**kwargs)
    output = kwargs["output"]
    with open(output, mode="w", encoding="utf-8") as file_handler:
        json.dump(geo.json, file_handler)


def cli_shape2ign(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]

    kwargs = parse_args_shape2ign(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ shape2ign ******************")
    catchment = kwargs.get("catchment")
    infile = kwargs.get("infile")
    output = kwargs.get("output")
    indent = kwargs.get("indent")
    ref_proj = kwargs.get("ref_proj")

    shape2ign(catchment, infile, output, ref_proj, indent=indent)


def soda(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]

    kwargs = parse_args_surfex_binary(argv, "soda")
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ soda ******************")
    run_surfex_binary("soda", **kwargs)


def titan(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]

    kwargs = parse_args_titan(argv)
    debug = kwargs.get("debug")

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ titan ******************")
    run_titan(**kwargs)



"""Pseuodo-obs."""

import logging

import numpy as np

from .interpolation import gridpos2points, inside_grid, sum_neighbour_points
from .netcdf import read_cryoclim_nc, read_sentinel_nc
from .obs import ObservationSet
from .observation import Observation


def snow_pseudo_obs_cryoclim(
    validtime,
    grid_snow_class,
    grid_lons,
    grid_lats,
    step,
    fg_geo,
    grid_snow_fg,
    gelevs_fg,
    grid_perm_snow=None,
    grid_perm_snow_geo=None,
    grid_slope=None,
    grid_slope_geo=None,
    fg_threshold=0.4,
    new_snow_depth=0.1,
    glaf=None,
    laf_threshold=0.1,
    neighbourhood=True,
):
    """Cryoclim snow.

    Args:
        validtime (as_datetime): Valid time
        grid_snow_class (np.ndarray): Snow class
        grid_lons (np.ndarray): Grid longitudes
        grid_lats (np.ndarray): Grid latitudes
        step (integer):  Step to process grid points.
        fg_geo (Geo): Geometry
        grid_snow_fg (np.ndarray): First guess snow
        gelevs_fg (np.ndarray): Grid elevations
        grid_perm_snow (np.ndarray): Permanent snow
        grid_perm_snow_geo (pysurfex.geo.Geo): Permant snow field geometry
        grid_slope (np.ndarray): Slope array
        grid_slope_geo (pysurfex.geo.Geo): Slope field geometry
        fg_threshold (float, optional): First guess threshold. Defaults to 2.0.
        new_snow_depth (float, optional): New snow depth. Defaults to 0.01.
        glaf (np.ndarray, optional): LandAreaFraction. Defaults to None
        laf_threshold(float): Threshold to remove points. Defaults to 0.1.
        neighbourhood(bool): Use neighbourhood. Default to True.

    Returns:
        list: List of observation objects

    """
    n_x = grid_lons.shape[0]
    n_y = grid_lons.shape[1]

    n_x = int(n_x / step)
    n_y = int(n_y / step)

    # TODO rewrite to use lonlatvals geo
    counter = 0
    iii = 0
    res_lons = []
    res_lats = []

    if neighbourhood:
        cryo_radius = 2
        neighbour_snow_cryo = grid_snow_class[0, :, :].copy()
        neighbour_snow_cryo[neighbour_snow_cryo != 2] = 0
        neighbour_snow_cryo[neighbour_snow_cryo == 2] = 1
        neighbour_snow_cryo = sum_neighbour_points(neighbour_snow_cryo, cryo_radius)
        neighbour_cryo_extra = grid_snow_class[0, :, :].copy()
        neighbour_cryo_extra[neighbour_cryo_extra == 4] = -4
        neighbour_cryo_extra[neighbour_cryo_extra >= 0] = 0
        neighbour_cryo_extra[neighbour_cryo_extra == -1] = 1
        neighbour_cryo_extra[neighbour_cryo_extra == -4] = 1
        neighbour_cryo_extra = sum_neighbour_points(neighbour_cryo_extra, cryo_radius)
    p_snow_class = {}
    p_neighbours = {}
    for __ in range(n_x):
        jjj = 0
        for __ in range(n_y):
            if (
                (
                    (grid_snow_class[0, iii, jjj] == 2)
                    or (grid_snow_class[0, iii, jjj] == 1)
                )
                and ((grid_lats[iii, jjj] < 90) and (grid_lats[iii, jjj] > -90.0))
                and ((grid_lons[iii, jjj] < 180) and (grid_lons[iii, jjj] > -180.0))
            ):
                res_lons.append(np.float64(grid_lons[iii, jjj]))
                res_lats.append(np.float64(grid_lats[iii, jjj]))
                p_snow_class.update({str(counter): grid_snow_class[0, iii, jjj]})
                if neighbourhood:
                    p_neighbours.update(
                        {
                            str(counter): neighbour_snow_cryo[iii, jjj]
                            + neighbour_cryo_extra[iii, jjj]
                        }
                    )
                counter = counter + 1
            jjj = jjj + step
        iii = iii + step

    logging.debug("len(grid_snow_fg)=%s", len(grid_snow_fg))
    p_fg_snow_depth = gridpos2points(
        fg_geo.lons, fg_geo.lats, np.asarray(res_lons), np.asarray(res_lats), grid_snow_fg
    )
    p_fg_elevs = gridpos2points(
        fg_geo.lons, fg_geo.lats, np.asarray(res_lons), np.asarray(res_lats), gelevs_fg
    )

    if neighbourhood:
        fg_snow_radius = 3
        p_fg_snow_depth_has_snow = grid_snow_fg
        p_fg_snow_depth_has_snow[p_fg_snow_depth_has_snow > 0] = 1
        p_fg_snow_depth_has_snow[p_fg_snow_depth_has_snow <= 0] = 0
        grid_having_snow = sum_neighbour_points(p_fg_snow_depth_has_snow, fg_snow_radius)
        points_having_snow = gridpos2points(
            fg_geo.lons,
            fg_geo.lats,
            np.asarray(res_lons),
            np.asarray(res_lats),
            grid_having_snow,
            operator="nearest",
        )

    if grid_perm_snow is not None:
        p_perm_snow = gridpos2points(
            grid_perm_snow_geo.lons,
            grid_perm_snow_geo.lats,
            np.asarray(res_lons),
            np.asarray(res_lats),
            grid_perm_snow,
        )

    if grid_slope is not None:
        p_slope = gridpos2points(
            grid_slope_geo.lons,
            grid_slope_geo.lats,
            np.asarray(res_lons),
            np.asarray(res_lats),
            grid_slope,
        )

    if glaf is not None:
        p_fg_laf = gridpos2points(
            fg_geo.lons, fg_geo.lats, np.asarray(res_lons), np.asarray(res_lats), glaf
        )
    logging.debug("fg_geo.lons=%s", fg_geo.lons)
    logging.debug("fg_geo.lats=%s", fg_geo.lats)
    logging.debug("res_lons=%s", res_lons)
    logging.debug("res_lats=%s", res_lats)
    in_grid = inside_grid(
        np.asarray(fg_geo.lons),
        np.asarray(fg_geo.lats),
        np.asarray(res_lons),
        np.asarray(res_lats),
        distance=2500.0,
    )

    # Ordering of points must be the same.....
    obs = []
    flags = []
    cis = []
    lafs = []
    providers = []
    logging.debug("p_fg_snow_depth.shape[0]=%s", p_fg_snow_depth.shape[0])
    for i in range(p_fg_snow_depth.shape[0]):
        p_snow_fg = p_fg_snow_depth[i]
        logging.debug("%s %s %s %s", i, p_snow_fg, res_lons[i], res_lats[i])
        if not np.isnan(p_snow_fg):
            laf_ok = True
            if glaf is not None and p_fg_laf[i] < laf_threshold:
                laf_ok = False
                logging.debug(
                    "Remove position because p_fg_laf=%s < %s",
                    p_fg_laf[i],
                    laf_threshold,
                )
            perm_ok = True
            if grid_perm_snow is not None and p_perm_snow[i] > 0.0:
                perm_ok = False

            slope_ok = True
            if grid_slope is not None and p_slope[i] > 0.5:
                slope_ok = False

            # Check if in grid
            if in_grid[i] and laf_ok and perm_ok and slope_ok:
                obs_value = np.nan
                try:
                    snow_class_val = p_snow_class[str(i)]
                except KeyError:
                    continue
                if snow_class_val == 2:
                    do = False
                    if neighbourhood:
                        if (
                            points_having_snow[i] < (fg_snow_radius * 2 + 1) ** 2
                            or p_neighbours[str(i)] < (cryo_radius * 2 + 1) ** 2
                        ):
                            do = True
                    else:
                        do = True
                    if do:
                        if p_snow_fg > 0:
                            if fg_threshold is not None:
                                if p_snow_fg <= fg_threshold:
                                    obs_value = p_snow_fg
                            else:
                                obs_value = p_snow_fg
                        else:
                            obs_value = new_snow_depth
                elif snow_class_val == 1:
                    if neighbourhood and points_having_snow[i] > 0:  # noqa SIM114
                        obs_value = 0.0
                    elif p_snow_fg >= 0.0:
                        obs_value = 0.0

                if not np.isnan(obs_value):
                    logging.debug("Add observation")
                    flags.append(0)
                    cis.append(0)
                    lafs.append(0)
                    providers.append(0)
                    obs.append(
                        Observation(
                            validtime,
                            res_lons[i],
                            res_lats[i],
                            obs_value,
                            p_fg_elevs[i],
                            varname="totalSnowDepth",
                        )
                    )
                else:
                    logging.debug("Value is nan")
            elif not in_grid[i]:
                logging.debug("Outside grid")

    logging.info("Possible pseudo-observations: %s", n_x * n_y)
    logging.info("Pseudo-observations created: %s", len(obs))
    return obs


def sm_obs_sentinel(
    validtime,
    grid_sm_class,
    grid_lons,
    grid_lats,
    step,
    fg_geo,
    grid_sm_fg,
    fg_threshold=1.0,
):
    """Sentinel.

    Args:
        validtime (as_datetime): Valid time
        grid_sm_class (np.ndarray): Soil moisture class
        grid_lons (np.ndarray): Grid longitudes
        grid_lats (np.ndarray): Grid latitudes
        step (integer):  Step to process grid points.
        fg_geo (Geo): Geometry
        grid_sm_fg (_type_): First guess
        fg_threshold (_type_, optional): First guess threshold. Defaults to 1..

    Returns:
        list: Observations

    """
    n_x = grid_lons.shape[0]
    n_y = grid_lons.shape[1]

    n_x = int(n_x / step)
    n_y = int(n_y / step)

    # TODO rewrite to use lonlatvals geo
    counter = 0
    iii = 0
    res_lons = []
    res_lats = []
    p_sm_class = {}
    for __ in range(n_x):
        jjj = 0
        for __ in range(n_y):
            res_lons.append(np.float64(grid_lons[iii, jjj]))
            res_lats.append(np.float64(grid_lats[iii, jjj]))
            p_sm_class.update({str(counter): grid_sm_class[iii, jjj]})
            counter = counter + 1
            jjj = jjj + step
        iii = iii + step

    p_fg_sm = gridpos2points(
        fg_geo.lons, fg_geo.lats, np.asarray(res_lons), np.asarray(res_lats), grid_sm_fg
    )
    in_grid = inside_grid(
        fg_geo.lons,
        fg_geo.lats,
        np.asarray(res_lons),
        np.asarray(res_lats),
        distance=2500.0,
    )
    # Ordering of points must be the same.....
    obs = []
    flags = []
    cis = []
    lafs = []
    providers = []
    for i in range(p_fg_sm.shape[0]):
        p_sm_fg = p_fg_sm[i]
        if not np.isnan(p_sm_fg) and in_grid[i]:
            obs_value = np.nan
            if (p_sm_class[str(i)] > 1) or (p_sm_class[str(i)] < 0):
                obs_value = 999
                if p_sm_fg <= fg_threshold:
                    obs_value = p_sm_fg
            else:
                obs_value = p_sm_class[str(i)]

            if not np.isnan(obs_value):
                flags.append(0)
                cis.append(0)
                lafs.append(0)
                providers.append(0)
                obs.append(
                    Observation(
                        validtime,
                        res_lons[i],
                        res_lats[i],
                        obs_value,
                        varname="surface_soil_moisture",
                    )
                )

    logging.info("Possible pseudo-observations: %s", n_x * n_y)
    logging.info("Pseudo-observations created: %s", len(obs))
    return obs


class SentinelObservationSet(ObservationSet):
    """JSON observation set."""

    def __init__(
        self,
        filename,
        validtime,
        fg_geo,
        grid_sm_fg,
        label="sentinel",
        step=2,
        fg_threshold=1.0,
    ):
        """Construct an observation data set from a json file.

        Args:
            filename (list): Filename
            validtime (as_datetime): Valdid time
            fg_geo (Geo): Surfex geometry
            grid_sm_fg (np.ndarray): Snow first guess field
            label (str, optional): Label of set. Defaults to "sentinel".
            step (int, optional): Step to process grid points. Defaults to 2.
            fg_threshold (float, optional): First guess threshold. Defaults to 1.0

        """
        grid_lons, grid_lats, grid_sm_class = read_sentinel_nc(filename)
        observations = sm_obs_sentinel(
            validtime,
            grid_sm_class,
            grid_lons,
            grid_lats,
            step,
            fg_geo,
            grid_sm_fg,
            fg_threshold,
        )

        ObservationSet.__init__(self, observations, label=label)


class CryoclimObservationSet(ObservationSet):
    """JSON observation set."""

    def __init__(
        self,
        filenames,
        validtime,
        fg_geo,
        snow_fg,
        gelevs_fg,
        perm_snow=None,
        perm_snow_geo=None,
        slope=None,
        slope_geo=None,
        label="cryo",
        step=2,
        fg_threshold=0.4,
        new_snow_depth=0.1,
        glaf=None,
        laf_threshold=0.1,
        cryo_varname="classed_value_c",
    ):
        """Construct an observation data set from a json file.

        Args:
            filenames (list): Filename
            validtime (as_datetime): Valdid time
            fg_geo (Geo): Surfex geometry
            snow_fg (np.ndarray): Snow first guess field
            gelevs_fg (np.ndarray): Grid elevations
            perm_snow (np.ndarray): Permanent snow
            perm_snow_geo (pysurfex.geo.Geo): Permant snow field geometry
            slope (np.ndarray): Slope array
            slope_geo (pysurfex.geo.Geo): Slope field geometry
            label (str, optional): Label of set. Defaults to "cryo".
            step (int, optional): Step to process grid points. Defaults to 2.
            fg_threshold (float, optional): First guess threshold. Defaults to 0.4
            new_snow_depth (float, optional): New snow depth in cryoclim in m.
                                              Defaults to 0.1
            glaf (np.ndarray, optional): Land-area-fraction. Defaults to None.
            laf_threshold (float, optional): Threshold for existing land-area-fraction.
                                             Defaults to 0.1.
            cryo_varname (str, optional): Variable name in cryo file.
                                          Defaults to "classed_value_c"

        """
        grid_lons, grid_lats, grid_snow_class = read_cryoclim_nc(
            filenames, cryo_varname=cryo_varname
        )
        observations = snow_pseudo_obs_cryoclim(
            validtime,
            grid_snow_class,
            grid_lons,
            grid_lats,
            step,
            fg_geo,
            snow_fg,
            gelevs_fg,
            fg_threshold=fg_threshold,
            new_snow_depth=new_snow_depth,
            grid_perm_snow=perm_snow,
            grid_perm_snow_geo=perm_snow_geo,
            grid_slope=slope,
            grid_slope_geo=slope_geo,
            glaf=glaf,
            laf_threshold=laf_threshold,
        )

        ObservationSet.__init__(self, observations, label=label)



#!/usr/bin/env python3
"""Implement helper routines to deal with dates and times."""
from datetime import date, datetime, timedelta, timezone


# TODO use ISO times
def as_datetime(dtg, offset=False):
    """Convert string to datetime."""
    if len(dtg) == 10:
        fmt = "%Y%m%d%H"
    elif len(dtg) == 12:
        fmt = "%Y%m%d%H%M"
    elif len(dtg) == 14:
        fmt = "%Y%m%d%H%M%S"
    else:
        raise RuntimeError(f"dtg={dtg} len(dtg) is {len(dtg)}")
    if offset:
        return datetime.strptime(dtg, fmt)
    return datetime.strptime(dtg, fmt).replace(tzinfo=timezone.utc)


def as_datetime_string(dtg):
    """Convert string to datetime."""
    fmt = "%Y%m%d%H%M%S"
    return dtg.strftime(fmt)


def offsetaware(dtg):
    """Make offset aware."""
    return dtg.replace(tzinfo=timezone.utc)


def as_timedelta(seconds=0):
    """Convert seconds to timedelta."""
    return timedelta(seconds=seconds)


def fromtimestamp(validtime):
    """Convert timestamp to validtime."""
    return datetime.fromtimestamp(validtime)


def utcfromtimestamp(epochtime):
    """Convert timestamp to validtime."""
    return datetime.utcfromtimestamp(epochtime).replace(tzinfo=timezone.utc)


def isdatetime(obj):
    """Check if is a datetime objects."""
    return isinstance(obj, date)


def as_datetime_args(year=None, month=None, day=None, hour=0, minute=0, second=0):
    """Set datetime object from args."""
    return datetime(
        year=year, month=month, day=day, hour=hour, minute=minute, second=second
    ).replace(tzinfo=timezone.utc)


def get_decade(dt) -> str:
    """Return the decade given a datetime object."""
    # Extract month and day from datetime object
    dtg_mm = int(dt.month)
    dtg_dd = int(dt.day)

    # Determine decades_mm and decades_dd based on dtg_dd
    if dtg_dd < 9:
        decades_mm = dtg_mm
        decades_dd = 5
    elif 8 < dtg_dd < 19:
        decades_mm = dtg_mm
        decades_dd = 15
    elif 18 < dtg_dd < 29:
        decades_mm = dtg_mm
        decades_dd = 25
    else:
        decades_mm = dtg_mm + 1
        if decades_mm == 13:
            decades_mm = 1
        decades_dd = 5

    decades_mm = f"{decades_mm:02d}"
    decades_dd = f"{decades_dd:02d}"

    return f"{decades_mm}{decades_dd}"
